from typing import Literal, TypedDict

from qtpy import QtGui, QtCore
from cmap import Color
from magicgui.types import Undefined
from magicgui.type_map import register_type
from magicgui.widgets import ComboBox, TupleEdit, Container
from magicgui.widgets.bases import ValuedContainerWidget

from himena.consts import StandardType, MenuId
from himena.exceptions import Cancelled
from himena.widgets import SubWindow, MainWindow
from himena.plugins import register_function, configure_gui
from himena.types import Parametric, Rect
from himena.standards import roi as _roi
from himena.utils import image_utils
from himena.qt.magicgui import ColorEdit, get_type_map, FloatEdit, ToggleSwitch
from himena import anchor

from .image import QImageView
from ._image_components._scale_bar import ScaleBarAnchor, ScaleBarType


### Commands specific to built-in widgets ###
@register_function(
    title="Scale bar ...",
    types=StandardType.IMAGE,
    menus=[MenuId.TOOLS_IMAGE],
    command_id="builtins:image:setup-image-scale-bar",
)
def setup_image_scale_bar(win: SubWindow[QImageView]) -> Parametric:
    """Set up the scale bar for the image viewer."""
    scale_bar = win.widget._img_view._scale_bar_widget

    @configure_gui(
        anchor={"value": scale_bar._anchor},
        type={"value": scale_bar._scale_bar_type},
        color={"widget_type": ColorEdit, "value": scale_bar._color.name()},
        preview=True,
    )
    def setup(
        visible: bool = True,
        text_visible: bool = True,
        anchor: ScaleBarAnchor = ScaleBarAnchor.BOTTOM_RIGHT,
        type: ScaleBarType = ScaleBarType.SHADOWED,
        color="white",
    ):
        qcolor = QtGui.QColor.fromRgbF(*Color(color).rgba)
        win.widget._img_view._scale_bar_widget.update_scale_bar(
            anchor=anchor, type=type, color=qcolor, visible=visible,
            text_visible=text_visible
        )  # fmt: skip

    return setup


@register_function(
    title="Set zoom factor ...",
    types=StandardType.IMAGE,
    menus=[MenuId.TOOLS_IMAGE],
    command_id="builtins:image:set-zoom-factor",
)
def set_zoom_factor(win: SubWindow[QImageView]) -> Parametric:
    """Set the zoom factor of the canvas."""
    view = win.widget._img_view
    current_scale = view.transform().m11()

    @configure_gui(
        scale={
            "value": round(current_scale * 100, 2),
            "min": 0.001,
            "label": "Zoom (%)",
        }
    )
    def run_set_zoom(scale: float):
        ratio = scale / (current_scale * 100)
        view.scale_and_update_handles(ratio)

    return run_set_zoom


@register_function(
    title="Capture Setting ...",
    types=StandardType.IMAGE,
    menus=[MenuId.TOOLS_IMAGE_CAPTURE, "/model_menu/capture"],
    command_id="builtins:image:capture-setting",
    group="00-capture",
)
def capture_setting(ui: MainWindow) -> Parametric:
    """Open the setting of image capture."""
    capture = ImageViewCapture.get_instance(ui)

    @configure_gui(
        auto_close=False,
        title="Capture Setting",
        scale_bars={"value": capture._scale_bars.copy(), "layout": "vertical"},
    )
    def run(
        scale_bars: list[ScaleBarSpecDict],
    ):
        capture._scale_bars = scale_bars.copy()
        return

    return run


@register_function(
    title="Copy slice to clipboard",
    types=StandardType.IMAGE,
    menus=[MenuId.TOOLS_IMAGE_CAPTURE, "/model_menu/capture"],
    command_id="builtins:image:copy-slice-to-clipboard",
    group="00-capture",
)
def copy_slice_to_clipboard(win: SubWindow[QImageView], ui: MainWindow):
    """Copy the current slice to the clipboard as is."""
    qimage = ImageViewCapture.get_instance(ui).to_qimage(win.widget)
    QtGui.QGuiApplication.clipboard().setImage(qimage)
    ui.show_notification("Image slice copied to the clipboard.", duration=3)


@register_function(
    title="Save slice to clipboard",
    types=StandardType.IMAGE,
    menus=[MenuId.TOOLS_IMAGE_CAPTURE, "/model_menu/capture"],
    command_id="builtins:image:save-slice",
    group="00-capture",
)
def save_slice(win: SubWindow[QImageView], ui: MainWindow):
    """Save the current slice to file as is."""
    if path := ui.exec_file_dialog(
        mode="w",
        extension_default=".png",
        allowed_extensions=[".png", ".jpg", ".jpeg"],
        caption="Save slice to clipboard",
    ):
        qimage = ImageViewCapture.get_instance(ui).to_qimage(win.widget)
        qimage.save(str(path))
    return None


@register_function(
    title="Copy viewer screenshot",
    types=StandardType.IMAGE,
    menus=[MenuId.TOOLS_IMAGE_CAPTURE, "/model_menu/capture"],
    command_id="builtins:image:copy-viewer-screenshot",
    group="10-screenshot",
)
def copy_image_view_screenshot(win: SubWindow[QImageView], ui: MainWindow):
    """Copy the screenshot of the image view to the clipboard."""
    qimage = win.widget._img_view.grab().toImage()
    QtGui.QGuiApplication.clipboard().setImage(qimage)
    ui.show_notification("Image view screenshot copied to the clipboard.", duration=3)


@register_function(
    title="Save viewer screenshot",
    types=StandardType.IMAGE,
    menus=[MenuId.TOOLS_IMAGE_CAPTURE, "/model_menu/capture"],
    command_id="builtins:image:save-viewer-screenshot",
    group="10-screenshot",
)
def save_image_view_screenshot(win: SubWindow[QImageView], ui: MainWindow):
    """Save the screenshot of the image view to a file."""
    qimage = win.widget._img_view.grab().toImage()

    if file_path := ui.exec_file_dialog(
        mode="w",
        extension_default=".png",
        allowed_extensions=[".png", ".jpg", ".jpeg"],
        caption="Save image view screenshot",
    ):
        return qimage.save(str(file_path))
    raise Cancelled


class ImageViewCapture:
    _instances: "dict[int, ImageViewCapture]" = {}

    def __init__(self):
        self._scale_bars: list[ScaleBarSpecDict] = []

    @classmethod
    def get_instance(cls, ui: MainWindow) -> "ImageViewCapture":
        if (self := cls._instances.get(id(ui))) is None:
            self = cls()
            cls._instances[id(ui)] = self
        return self

    def to_qimage(self, view: QImageView) -> QtGui.QImage:
        if isinstance(current_roi := view.current_roi(), _roi.RectangleRoi):
            bbox = image_utils.roi_2d_to_bbox(current_roi, view._arr, view._is_rgb)
        else:
            bbox = Rect(0, 0, view._arr.shape[-1], view._arr.shape[-2])
        # manually paint the graphics items
        qimage = QtGui.QImage(
            bbox.width, bbox.height, QtGui.QImage.Format.Format_RGBA8888
        )
        painter = QtGui.QPainter(qimage)
        target_rect = QtCore.QRect(0, 0, bbox.width, bbox.height)
        source_rect = QtCore.QRect(bbox.left, bbox.top, bbox.width, bbox.height)
        for graphics in view._img_view._image_widgets:
            if graphics.isVisible():
                graphics.initPainter(painter)
                painter.drawImage(target_rect, graphics._qimage.copy(source_rect))

        # draw scale bars
        painter.setPen(QtCore.Qt.PenStyle.NoPen)
        for sbar in self._scale_bars:
            width, height = sbar["shape"]
            axes = view.dims_slider.to_dim_axes()
            w_scale, h_scale = axes[-1].scale, axes[-2].scale
            w_px = int(round(width / w_scale))
            h_px = int(round(height / h_scale))
            qcolor = QtGui.QColor.fromRgbF(*Color(sbar["color"]).rgba)
            if rect := to_anchor_object(sbar, w_px, h_px).apply_anchor(
                bbox.size(), (w_px, h_px)
            ):
                painter.setBrush(qcolor)
                painter.drawRect(rect.left, rect.top, rect.width, rect.height)

        painter.end()
        return qimage


class ScaleBarSpecDict(TypedDict):
    shape: tuple[float, float]
    color: str
    anchor_pos: Literal["top-left", "top-right", "bottom-left", "bottom-right"]
    offset: "tuple[int, int] | None" = None


def to_anchor_object(
    sdict: ScaleBarSpecDict,
    w_px: int,
    h_px: int,
) -> anchor.WindowAnchor:
    anc_cls = anchor.type_to_anchor_class(sdict["anchor_pos"] + "-const")
    offset = sdict.get("offset")
    if offset is None:
        offset = (max(min(w_px, h_px), 1),) * 2
    return anc_cls(*offset)


class ScaleBarSpecWidget(ValuedContainerWidget[ScaleBarSpecDict]):
    def __init__(self, value=Undefined, **kwargs):
        self._width = FloatEdit(
            value=5,
            min=0,
            label="width",
            tooltip="Scale bar width in the unit of the image scale",
        )
        self._height = FloatEdit(
            value=0.5,
            min=0,
            label="height",
            tooltip="Scale bar height in the unit of the image scale",
        )
        self._color = ColorEdit(
            "white", label="color", tooltip="color of the scale bar"
        )
        self._anchor_pos = ComboBox(
            value="bottom-right",
            choices=["top-left", "top-right", "bottom-left", "bottom-right"],
            label="anchor",
        )
        self._use_offset = ToggleSwitch(value=False, text="Constant offset")
        self._offset = TupleEdit(value=(10, 10), label="offset (pixels)")
        c0 = Container(widgets=[self._width, self._height], layout="horizontal")
        c1 = Container(widgets=[self._anchor_pos, self._color], layout="horizontal")
        c2 = Container(widgets=[self._offset, self._use_offset], layout="horizontal")
        c0.margins = c1.margins = c2.margins = (0, 0, 0, 0)
        super().__init__(
            value=value,
            widgets=[c0, c1, c2],
            **kwargs,
        )
        for wd in [
            self._width,
            self._height,
            self._color,
            self._anchor_pos,
            self._use_offset,
            self._offset,
        ]:
            wd.changed.connect(self._on_changed)
        self._use_offset.changed.connect(self._use_offset_changed)
        self._use_offset_changed(self._use_offset.value)

    def _use_offset_changed(self, use_offset: bool):
        self._offset.enabled = use_offset

    def _on_changed(self):
        self.changed.emit(self.get_value())

    def get_value(self):
        return {
            "shape": (self._width.value, self._height.value),
            "color": self._color.get_value().hex,
            "anchor_pos": self._anchor_pos.value,
            "offset": self._offset.value if self._use_offset.value else None,
        }

    def set_value(self, val):
        if val == Undefined:
            return
        val = ScaleBarSpecDict(**val)
        self._width.value, self._height.value = val["shape"]
        self._color.value = Color(val["color"]).hex
        self._anchor_pos.value = val["anchor_pos"]
        self._use_offset.value = val["offset"] is not None
        if self._use_offset.value:
            self._offset.value = val["offset"]


type_map = get_type_map()
type_map.register_type(ScaleBarSpecDict, widget_type=ScaleBarSpecWidget)
register_type(ScaleBarSpecDict, widget_type=ScaleBarSpecWidget)



from __future__ import annotations
import weakref

from qtpy import QtWidgets as QtW
from qtpy import QtGui, QtCore

from himena.consts import StandardType, MonospaceFontFamily
from himena.qt import drag_model, ndarray_to_qimage
from himena import _drag
from himena.standards.model_meta import TextMeta
from himena.standards import ipynb
from himena.style import Theme
from himena.types import WidgetDataModel
from himena.plugins import validate_protocol

from himena_builtins.qt.widgets._text_base import QMainTextEdit
from himena_builtins.qt.widgets._dragarea import QDraggableArea


class QIpynbEdit(QtW.QScrollArea):
    """The built-in ipynb (Jupyter Notebook) editor widget.

    ## Basic Usage

    This widget is only for editing Jupyter Notebook files. It can add or delete cells,
    edit cells and move cells. Running the code is not supported.

    ## Drag and Drop

    Each cell can be dragged out using the drag indicator. The dragged data has type
    `StandardType.TEXT` ("text"). If the data is dropped in the same widget, the cell
    will be moved, otherwise the cell content will be copied.
    """

    __himena_widget_id__ = "builtins:QIpynbEdit"
    __himena_display_name__ = "Built-in Jupyter Notebook Editor"

    def __init__(self):
        super().__init__()
        self._central_widget = QtW.QWidget()
        self.setWidget(self._central_widget)
        self.setWidgetResizable(True)
        self.setHorizontalScrollBarPolicy(QtCore.Qt.ScrollBarPolicy.ScrollBarAlwaysOff)

        self._cell_widgets: list[QIpynbCellEdit] = []
        _layout = QtW.QVBoxLayout(self._central_widget)
        _layout.setContentsMargins(2, 2, 2, 2)
        _layout.setAlignment(QtCore.Qt.AlignmentFlag.AlignTop)
        self._layout = _layout
        self._ipynb_orig = ipynb.IpynbFile()
        self._model_type = StandardType.IPYNB
        self._dragging_index: int | None = None
        self._text_theme = "default"
        self._control_widget = QIpynbControl(self)
        self._is_editable = True

    @validate_protocol
    def update_model(self, model: WidgetDataModel):
        if not isinstance(value := model.value, str):
            value = str(value)
        self._ipynb_orig = ipynb.IpynbFile.model_validate_json(value)
        self.clear_all()
        for idx, cell in enumerate(self._ipynb_orig.cells):
            self.insert_cell(idx, cell)
        self._model_type = model.type
        return None

    @validate_protocol
    def to_model(self) -> WidgetDataModel:
        ipynb = self._ipynb_orig.model_copy()
        for idx, widget in enumerate(self._cell_widgets):
            ipynb.cells[idx].source = widget.text()
        js_string = ipynb.model_dump_json(indent=2)
        return WidgetDataModel(
            type=self.model_type(),
            value=js_string,
            extension_default=".ipynb",
        )

    @validate_protocol
    def model_type(self) -> str:
        return self._model_type

    @validate_protocol
    def is_modified(self) -> bool:
        return any(
            widget._text_edit.isWindowModified() for widget in self._cell_widgets
        )

    @validate_protocol
    def theme_changed_callback(self, theme: Theme):
        if theme.is_light_background():
            self._text_theme = "default"
        else:
            self._text_theme = "native"
        for widget in self._cell_widgets:
            widget._text_edit._code_theme = self._text_theme
            widget._text_edit.syntax_highlight(widget._text_edit._language)

    @validate_protocol
    def size_hint(self) -> tuple[int, int]:
        return 400, 360

    @validate_protocol
    def is_editable(self) -> bool:
        return self._is_editable

    @validate_protocol
    def set_editable(self, editable: bool):
        self._is_editable = editable
        for widget in self._cell_widgets:
            widget._text_edit.setReadOnly(not editable)
        self._control_widget._insert_cell_btn.setEnabled(editable)
        self._control_widget._insert_md_btn.setEnabled(editable)
        self._control_widget._delete_cell_btn.setEnabled(editable)

    @validate_protocol
    def control_widget(self) -> QIpynbControl:
        return self._control_widget

    def clear_all(self):
        for child in self._cell_widgets:
            self._layout.removeWidget(child)
            child.deleteLater()
        self._cell_widgets.clear()

    def insert_cell(self, idx: int, cell: ipynb.IpynbCell):
        widget = QIpynbCellEdit(cell, self._ipynb_orig.language, self)
        self._layout.insertWidget(idx, widget)
        self._cell_widgets.insert(idx, widget)

    def delete_cell(self, idx: int):
        widget = self._cell_widgets.pop(idx)
        self._layout.removeWidget(widget)
        widget.deleteLater()

    def current_index(self) -> int:
        for idx, widget in enumerate(self._cell_widgets):
            if widget._text_edit.hasFocus() or widget.hasFocus():
                return idx
        return -1


class QIpynbCellEdit(QtW.QGroupBox):
    """Widget for a single cell."""

    def __init__(self, cell: ipynb.IpynbCell, language: str, parent: QIpynbEdit):
        super().__init__(parent)
        self._ipynb_cell = cell
        self.setAcceptDrops(True)
        self._ipynb_edit_ref = weakref.ref(parent)
        self._text_edit = QMainTextEdit()
        self._text_edit._code_theme = parent._text_theme
        self._text_edit.setPlainText(cell.source)
        self._text_edit.setSizePolicy(
            QtW.QSizePolicy.Policy.Expanding, QtW.QSizePolicy.Policy.Expanding
        )
        lang = language if cell.cell_type == "code" else cell.cell_type
        self._text_edit.syntax_highlight(lang)
        self._language_label = QtW.QLabel(lang.title())
        font = QtGui.QFont(MonospaceFontFamily)
        font.setPointSize(8)
        self._language_label.setFont(font)
        self._language_label.setFixedHeight(14)
        self._language_label.setAlignment(
            QtCore.Qt.AlignmentFlag.AlignRight | QtCore.Qt.AlignmentFlag.AlignVCenter
        )
        self._draggable_area = QDraggableArea()
        self._draggable_area.setFixedSize(14, 20)
        self._draggable_area.dragged.connect(self._drag_event)

        _layout = QtW.QVBoxLayout(self)
        _layout.setContentsMargins(2, 2, 2, 2)
        _layout.setSpacing(1)
        _layout.addWidget(self._text_edit)

        _footer_layout = QtW.QHBoxLayout()
        _footer_layout.setContentsMargins(1, 1, 1, 1)
        _footer_layout.addWidget(self._draggable_area)
        _footer_layout.addWidget(self._language_label)
        _layout.addLayout(_footer_layout)

        self._output_widget = None
        # NOTE: It seems that the byte content of QImage will be garbage collected
        # if we don't keep a reference to it.
        self._qimages: list[QtGui.QImage] = []
        if cell.outputs:
            output_widget = QIpynbOutput()
            _layout.addWidget(output_widget)
            for output in cell.outputs:
                if isinstance(output, ipynb.IpynbStreamOutput):
                    output_widget.append_text(output.get_text_plain())
                elif isinstance(output, ipynb.IpynbErrorOutput):
                    output_widget.append_html(output.get_html())
                elif isinstance(output, ipynb.IpynbDisplayDataOutput):
                    if (img := output.get_image()) is not None:
                        qimg = ndarray_to_qimage(img)
                        output_widget.append_image(qimg)
                        self._qimages.append(qimg)
                    elif html := output.get_text_html():
                        output_widget.append_html(html)
                    elif text := output.get_text_plain():
                        output_widget.append_text(text)

            self._output_widget = output_widget

        self._text_edit.textChanged.connect(self._on_text_changed)
        self._height_for_font = self._text_edit.fontMetrics().height()

    def _on_text_changed(self):
        nblocks = self._text_edit.blockCount()
        height = (self._height_for_font + 4) * min(nblocks, 10)
        if self._output_widget:
            height += self._output_widget.height()
        self.setFixedHeight(height + 40)

    def text(self) -> str:
        return self._text_edit.toPlainText()

    def _drag_event(self):
        cursor = self._text_edit.textCursor()
        font = self._text_edit.font()
        model = WidgetDataModel(
            value=self.text(),
            type=StandardType.TEXT,
            metadata=TextMeta(
                language=self._language_label.text().lower(),
                selection=(cursor.selectionStart(), cursor.selectionEnd()),
                font_family=font.family(),
                font_size=font.pointSizeF(),
            ),
        )
        ipynb_edit = self._ipynb_edit_ref()
        if ipynb_edit is not None:
            ipynb_edit._dragging_index = ipynb_edit._cell_widgets.index(self)
        drag_model(model, desc="Cell", source=ipynb_edit, text_data=model.value)

    def dragEnterEvent(self, a0):
        ipynb_edit = self._ipynb_edit_ref()
        if ipynb_edit is None or not ipynb_edit.is_editable():
            a0.ignore()
            return
        if model := _drag.get_dragging_model():
            if model.type == StandardType.TEXT:
                a0.accept()
                return
        a0.ignore()
        return

    def dropEvent(self, a0):
        if model := _drag.drop():
            ipynb_edit = self._ipynb_edit_ref()
            insert_idx = ipynb_edit._cell_widgets.index(self)
            if a0.pos().y() > self.rect().center().y():
                insert_idx += 1
            data_model = model.data_model()
            if a0.source() is ipynb_edit and ipynb_edit._dragging_index is not None:
                # move cell
                ipynb_edit.delete_cell(ipynb_edit._dragging_index)
                outputs = ipynb_edit._cell_widgets[insert_idx]._ipynb_cell.outputs
                if ipynb_edit._dragging_index < insert_idx:
                    insert_idx -= 1
                cell = ipynb.IpynbCell(
                    source=data_model.value,
                    cell_type="code",
                    outputs=outputs.copy(),
                )
            else:
                cell = ipynb.IpynbCell(source=data_model.value, cell_type="code")
            ipynb_edit.insert_cell(insert_idx, cell)


class QIpynbControl(QtW.QWidget):
    def __init__(self, parent: QIpynbEdit):
        super().__init__(parent)
        self._ipynb_edit = parent
        self._insert_cell_btn = QtW.QPushButton("+ Code")
        self._insert_cell_btn.clicked.connect(self._insert_code)
        self._insert_cell_btn.setFocusPolicy(QtCore.Qt.FocusPolicy.NoFocus)
        self._insert_md_btn = QtW.QPushButton("+ Markdown")
        self._insert_md_btn.clicked.connect(self._insert_md)
        self._insert_md_btn.setFocusPolicy(QtCore.Qt.FocusPolicy.NoFocus)
        self._delete_cell_btn = QtW.QPushButton("Delete")
        self._delete_cell_btn.clicked.connect(self._delete_cell)
        self._delete_cell_btn.setFocusPolicy(QtCore.Qt.FocusPolicy.NoFocus)
        self._clear_outputs_btn = QtW.QPushButton("Clear Outputs")
        self._clear_outputs_btn.clicked.connect(self._clear_outputs)
        self._clear_outputs_btn.setFocusPolicy(QtCore.Qt.FocusPolicy.NoFocus)

        layout = QtW.QHBoxLayout(self)
        layout.setContentsMargins(0, 0, 0, 0)
        layout.setAlignment(QtCore.Qt.AlignmentFlag.AlignRight)
        layout.addWidget(QtW.QWidget())  # spacer
        layout.addWidget(self._insert_cell_btn)
        layout.addWidget(self._insert_md_btn)
        layout.addWidget(self._delete_cell_btn)

    def _insert_code(self):
        return self._insert_impl("code")

    def _insert_md(self):
        return self._insert_impl("markdown")

    def _delete_cell(self):
        idx = self._ipynb_edit.current_index()
        if idx != -1:
            self._ipynb_edit.delete_cell(idx)

    def _insert_impl(self, cell_type: str):
        idx = self._ipynb_edit.current_index()
        if idx == -1:
            idx = len(self._ipynb_edit._cell_widgets) - 1
        cell = ipynb.IpynbCell(cell_type=cell_type)
        self._ipynb_edit.insert_cell(idx + 1, cell)

    def _clear_outputs(self):
        for widget in self._ipynb_edit._cell_widgets:
            if widget._output_widget is not None:
                widget._output_widget.clear()
                widget._qimages.clear()
                self._ipynb_edit._layout.removeWidget(widget)
                widget.deleteLater()
            widget._ipynb_cell.outputs = []


class QIpynbOutput(QtW.QTextEdit):
    def __init__(self):
        super().__init__()
        self.setReadOnly(True)
        self.setWordWrapMode(QtGui.QTextOption.WrapMode.NoWrap)
        self.setStyleSheet(
            "QIpynbOutput { border: 1px solid gray; margin-left: 10px; }"
        )
        self.setContextMenuPolicy(QtCore.Qt.ContextMenuPolicy.CustomContextMenu)
        font = QtGui.QFont(MonospaceFontFamily, 10)
        self.setFont(font)
        self.setFixedHeight(0)
        self._max_height = 400

        @self.customContextMenuRequested.connect
        def rightClickContextMenu(point):
            menu = self._make_contextmenu(point)
            if menu:
                menu.exec(self.mapToGlobal(point))

    def append_text(self, text: str):
        self.moveCursor(QtGui.QTextCursor.MoveOperation.End)
        self.insertPlainText(text)
        self.moveCursor(QtGui.QTextCursor.MoveOperation.End)
        if self.height() < self._max_height:
            font_height = self.fontMetrics().height() + 8
            new_height = self.height() + len(text.splitlines()) * font_height
            self.setFixedHeight(min(new_height, self._max_height))

    def append_html(self, html: str):
        self.moveCursor(QtGui.QTextCursor.MoveOperation.End)
        self.insertHtml(html)
        self.moveCursor(QtGui.QTextCursor.MoveOperation.End)
        if self.height() < self._max_height:
            font_height = self.fontMetrics().height() + 8
            new_height = self.height() + html.count("<br>") * font_height
            self.setFixedHeight(min(new_height, self._max_height))

    def append_image(self, qimage: QtGui.QImage):
        self.moveCursor(QtGui.QTextCursor.MoveOperation.End)
        cursor = self.textCursor()
        cursor.insertImage(qimage)
        self.insertPlainText("\n")
        self.moveCursor(QtGui.QTextCursor.MoveOperation.End)
        if self.height() < self._max_height:
            new_height = self.height() + qimage.height() + 16
            self.setFixedHeight(min(new_height, self._max_height))

    def _make_contextmenu(self, pos: QtCore.QPoint):
        """Reimplemented to return a custom context menu for images."""
        format = self.cursorForPosition(pos).charFormat()
        if name := format.stringProperty(QtGui.QTextFormat.Property.ImageName):
            menu = QtW.QMenu(self)
            menu.addAction("Copy Image", lambda: self._copy_image(name))
            menu.addAction("Save Image As...", lambda: self._save_image(name))
            return menu
        return None

    def _copy_image(self, name):
        image = self._get_image(name)
        if image is None:
            raise ValueError("Image not found")
        return QtW.QApplication.clipboard().setImage(image)

    def _save_image(self, name):
        """Shows a save dialog for the ImageResource with 'name'."""
        image = self._get_image(name)
        if image is None:
            raise ValueError("Image not found")
        dialog = QtW.QFileDialog(self, "Save Image")
        dialog.setAcceptMode(QtW.QFileDialog.AcceptMode.AcceptSave)
        dialog.setDefaultSuffix("png")
        dialog.setNameFilter("PNG file (*.png)")
        if dialog.exec_():
            filename = dialog.selectedFiles()[0]
            image.save(filename, "PNG")
        return None

    def _get_image(self, name: str) -> QtGui.QImage | None:
        """Returns the QImage stored as the ImageResource with 'name'."""
        if document := self.document():
            image = document.resource(
                QtGui.QTextDocument.ResourceType.ImageResource, QtCore.QUrl(name)
            )
            return image
        return None



import sys
from qtpy import QtWidgets as QtW
from qtpy.QtCore import Qt
from pytestqt.qtbot import QtBot
from superqt.utils import thread_worker
from himena import create_model, StandardType
from himena.qt._qtraceback import QtErrorMessageBox, QtTracebackDialog
from himena.qt import MainWindowQt
from himena.qt._qmodeldrop import QModelDrop, QModelDropList

def test_qt_traceback(qtbot: QtBot):
    from himena.qt._qtraceback import format_exc_info_py310, format_exc_info_py311

    exception = ValueError("Test value error")
    widget = QtW.QWidget()
    msgbox = QtErrorMessageBox("Test", exception, parent=widget)
    qtbot.addWidget(widget)
    qtbot.addWidget(msgbox)
    tb = msgbox._get_traceback()
    tb_dlg = QtTracebackDialog(msgbox)
    qtbot.addWidget(tb_dlg)
    tb_dlg.setText(tb)

    msgbox._traceback_button_clicked(runtime=False)

    if sys.version_info < (3, 11):
        format_exc_info_py310(msgbox._exc_info(), as_html=True)
    else:
        format_exc_info_py311(msgbox._exc_info(), as_html=True)

def test_tab_widget(qtbot: QtBot):
    from himena.qt._qtab_widget import QTabWidget

    tab_widget = QTabWidget()
    tab_widget.show()  # this is necessary for testing key click
    qtbot.addWidget(tab_widget)
    tab_widget.add_tab_area("X")
    tab_widget._line_edit.start_edit(0)
    QtW.QApplication.processEvents()
    tab_widget._line_edit.setText("Y")
    QtW.QApplication.processEvents()
    qtbot.keyClick(tab_widget._line_edit, Qt.Key.Key_Return)
    QtW.QApplication.processEvents()
    assert tab_widget.tabText(0) == "Y"

def test_int_line_edit(qtbot: QtBot):
    from himena.qt._qlineedit import QIntLineEdit

    line = QIntLineEdit()
    qtbot.addWidget(line)
    assert line.text() == ""

    # nothing changes
    qtbot.keyClick(line, Qt.Key.Key_Up)
    qtbot.keyClick(line, Qt.Key.Key_Down)
    qtbot.keyClick(line, Qt.Key.Key_PageUp)
    qtbot.keyClick(line, Qt.Key.Key_PageDown)

    qtbot.keyClick(line, Qt.Key.Key_A)
    assert line.text() == ""  # validator
    qtbot.keyClick(line, Qt.Key.Key_5)
    assert line.text() == "5"
    qtbot.keyClick(line, Qt.Key.Key_Up)
    assert line.text() == "6"
    qtbot.keyClick(line, Qt.Key.Key_Down)
    assert line.text() == "5"
    qtbot.keyClick(line, Qt.Key.Key_PageUp)
    assert line.text() == "105"
    qtbot.keyClick(line, Qt.Key.Key_PageDown)
    assert line.text() == "5"
    line.setText("")
    line._on_editing_finished()
    line._on_text_changed("")
    line._on_text_changed("0")

def test_double_line_edit(qtbot: QtBot):
    from himena.qt._qlineedit import QDoubleLineEdit

    line = QDoubleLineEdit()
    qtbot.addWidget(line)
    assert line.text() == ""

    # nothing changes
    qtbot.keyClick(line, Qt.Key.Key_Up)
    qtbot.keyClick(line, Qt.Key.Key_Down)
    qtbot.keyClick(line, Qt.Key.Key_PageUp)
    qtbot.keyClick(line, Qt.Key.Key_PageDown)

    qtbot.keyClick(line, Qt.Key.Key_A)
    assert line.text() == ""  # validator
    qtbot.keyClick(line, Qt.Key.Key_3)
    qtbot.keyClick(line, Qt.Key.Key_Period)
    qtbot.keyClick(line, Qt.Key.Key_1)
    assert line.text() == "3.1"
    qtbot.keyClick(line, Qt.Key.Key_Up)
    assert line.text() == "3.2"
    qtbot.keyClick(line, Qt.Key.Key_Down)
    assert line.text() == "3.1"
    qtbot.keyClick(line, Qt.Key.Key_PageUp)
    assert line.text() == "13.1"
    qtbot.keyClick(line, Qt.Key.Key_PageDown)
    assert line.text() == "3.1"
    line.setText("")
    line._on_editing_finished()
    line._on_text_changed("")
    line._on_text_changed("0")

def test_double_line_edit_exponential(qtbot: QtBot):
    from himena.qt._qlineedit import QDoubleLineEdit

    line = QDoubleLineEdit()
    qtbot.addWidget(line)
    assert line.text() == ""

    qtbot.keyClick(line, Qt.Key.Key_3)
    qtbot.keyClick(line, Qt.Key.Key_Period)
    qtbot.keyClick(line, Qt.Key.Key_1)
    qtbot.keyClick(line, Qt.Key.Key_E)
    qtbot.keyClick(line, Qt.Key.Key_2)
    assert line.text() == "3.1e2"
    qtbot.keyClick(line, Qt.Key.Key_Up)
    assert line.text() == "3.2e2"
    qtbot.keyClick(line, Qt.Key.Key_Down)
    assert line.text() == "3.1e2"
    qtbot.keyClick(line, Qt.Key.Key_PageUp)
    assert line.text() == "3.1e3"
    qtbot.keyClick(line, Qt.Key.Key_PageDown)
    assert line.text() == "3.1e2"

def test_int_list_line_edit(qtbot: QtBot):
    from himena.qt._qlineedit import QCommaSeparatedIntLineEdit

    line = QCommaSeparatedIntLineEdit()
    qtbot.addWidget(line)
    assert line.text() == ""

    qtbot.keyClick(line, Qt.Key.Key_1)
    qtbot.keyClick(line, Qt.Key.Key_Comma)
    assert line.text() == "1,"
    qtbot.keyClick(line, Qt.Key.Key_2)
    qtbot.keyClick(line, Qt.Key.Key_Comma)
    qtbot.keyClick(line, Qt.Key.Key_Space)
    qtbot.keyClick(line, Qt.Key.Key_3)
    assert line.text() == "1,2, 3"
    qtbot.keyClick(line, Qt.Key.Key_Comma)
    qtbot.keyClick(line, Qt.Key.Key_Comma)
    assert line.text() == "1,2, 3,"

def test_float_list_line_edit(qtbot: QtBot):
    from himena.qt._qlineedit import QCommaSeparatedDoubleLineEdit

    line = QCommaSeparatedDoubleLineEdit()
    qtbot.addWidget(line)
    assert line.text() == ""

    qtbot.keyClick(line, Qt.Key.Key_1)
    qtbot.keyClick(line, Qt.Key.Key_Comma)
    assert line.text() == "1,"
    qtbot.keyClick(line, Qt.Key.Key_2)
    qtbot.keyClick(line, Qt.Key.Key_Comma)
    qtbot.keyClick(line, Qt.Key.Key_Space)
    qtbot.keyClick(line, Qt.Key.Key_3)
    assert line.text() == "1,2, 3"
    qtbot.keyClick(line, Qt.Key.Key_Comma)
    qtbot.keyClick(line, Qt.Key.Key_Comma)
    assert line.text() == "1,2, 3,"

def test_int_line_edit_range(qtbot: QtBot):
    from himena.qt._qlineedit import QIntLineEdit

    line = QIntLineEdit()
    qtbot.addWidget(line)
    line.setMinimum(10)
    line.setMaximum(100)
    line.setText("10")
    assert line.text() == "10"
    line.stepUp()
    assert line.text() == "11"
    line.stepDown()
    assert line.text() == "10"
    line.stepDown()
    assert line.text() == "10"
    # check upper limit
    line.setText("100")
    line.stepDown()
    assert line.text() == "99"
    line.stepUp()
    assert line.text() == "100"
    line.stepUp()
    assert line.text() == "100"

def test_float_line_edit_range(qtbot: QtBot):
    from himena.qt._qlineedit import QDoubleLineEdit

    line = QDoubleLineEdit()
    qtbot.addWidget(line)

    line = QDoubleLineEdit()
    qtbot.addWidget(line)
    line.setMinimum(0.0)
    line.setMaximum(1.0)
    line.setText("0.0")
    assert line.text() == "0.0"
    line.stepUp()
    assert line.text() == "0.1"
    line.stepDown()
    assert line.text() == "0.0"
    line.stepDown()
    assert line.text() == "0.0"
    # check upper limit
    line.setText("1.0")
    line.stepDown()
    assert line.text() == "0.9"
    line.stepUp()
    assert line.text() == "1.0"
    line.stepUp()
    assert line.text() == "1.0"

def test_model_drop(himena_ui: MainWindowQt):
    himena_ui.add_object("abc", type=StandardType.TEXT)
    himena_ui.add_object([[0, 1], [2, 3]], type="table")
    qdrop = QModelDrop([StandardType.TEXT])
    qdroplist = QModelDropList([StandardType.TEXT])
    drop_win = himena_ui.add_widget(qdrop)
    drop_list = himena_ui.add_widget(qdroplist)
    qwins = himena_ui._backend_main_window._tab_widget.widget_area(0).subWindowList()
    qdrop._drop_qsubwindow(qwins[0])
    assert qdrop.to_model().value == "abc"
    assert qdrop.to_model().type == StandardType.TEXT
    qdrop._drop_qsubwindow(qwins[1])
    assert qdrop.to_model().value == "abc"  # type not accepted, so not changed
    assert qdrop.to_model().type == StandardType.TEXT
    qdrop.set_model(create_model("pqr", type=StandardType.TEXT))
    assert qdrop.to_model().value == "pqr"
    qdroplist._drop_qsubwindow(qwins[0])
    qdroplist._drop_qsubwindow(qwins[1])
    assert len(qdroplist.models()) == 1
    assert qdroplist.models()[0].type == StandardType.TEXT
    qdroplist.itemWidget(qdroplist.item(0))._update_btn_pos()
    qdroplist.set_models(None)

def test_notification(himena_ui: MainWindowQt):
    himena_ui.add_object("abc", type=StandardType.TEXT)

    @thread_worker
    def func():
        yield

    worker = func()
    himena_ui._backend_main_window._job_stack.add_worker(worker, "Test Job")
    worker.start()
    worker.await_workers(100)



from himena.qt.magicgui._register import register_magicgui_types, get_type_map
from himena.qt.magicgui._color import ColorEdit, ColormapEdit
from himena.qt.magicgui._plot_elements import (
    FacePropertyEdit,
    EdgePropertyEdit,
    AxisPropertyEdit,
    DictEdit,
)
from himena.qt.magicgui._toggle_switch import ToggleSwitch
from himena.qt.magicgui._basic_widgets import (
    IntEdit,
    FloatEdit,
    IntListEdit,
    FloatListEdit,
)
from himena.qt.magicgui._file_edit import FileEdit
from himena.qt.magicgui._selection import SelectionEdit
from himena.qt.magicgui._modeldrop import (
    ModelDrop,
    ModelListDrop,
    SubWindowDrop,
    SubWindowListDrop,
)
from himena.qt.magicgui._dtypeedit import NumericDTypeEdit
from himena.qt.magicgui._value_getter import SliderRangeGetter

__all__ = [
    "get_type_map",
    "register_magicgui_types",
    "ColorEdit",
    "ColormapEdit",
    "ToggleSwitch",
    "FacePropertyEdit",
    "EdgePropertyEdit",
    "FileEdit",
    "AxisPropertyEdit",
    "IntEdit",
    "FloatEdit",
    "IntListEdit",
    "FloatListEdit",
    "DictEdit",
    "ModelDrop",
    "ModelListDrop",
    "SubWindowDrop",
    "SubWindowListDrop",
    "NumericDTypeEdit",
    "SelectionEdit",
    "NumericDTypeEdit",
    "SliderRangeGetter",
]



from __future__ import annotations

import math
from typing import Any, Union
import numpy as np
from numpy.typing import NDArray
from pydantic import Field, field_validator
from himena.types import Rect
from himena.standards.roi._base import Roi1D, Roi2D
from himena.standards.roi import _utils

Scalar = Union[int, float]


class SpanRoi(Roi1D):
    """ROI that represents a span in 1D space."""

    start: float = Field(..., description="Start of the span.")
    end: float = Field(..., description="End of the span.")

    def shifted(self, dx: float) -> SpanRoi:
        return SpanRoi(start=self.start + dx, end=self.end + dx)

    def width(self) -> float:
        return self.end - self.start


class PointRoi1D(Roi1D):
    """ROI that represents a point in 1D space."""

    x: float = Field(..., description="X-coordinate of the point.")

    def shifted(self, dx: float) -> PointRoi1D:
        return PointRoi1D(x=self.x + dx)


class PointsRoi1D(Roi1D):
    """ROI that represents a set of points in 1D space."""

    xs: Any = Field(..., description="List of x-coordinates.")

    @field_validator("xs")
    def _validate_np_array(cls, v) -> NDArray[np.number]:
        out = np.asarray(v)
        if out.dtype.kind not in "if":
            raise ValueError("Must be a numerical array.")
        return out

    def shifted(self, dx: float) -> PointsRoi1D:
        return PointsRoi1D(xs=self.xs + dx)


class RectangleRoi(Roi2D):
    """ROI that represents a rectangle."""

    x: Scalar = Field(..., description="X-coordinate of the top-left corner.")
    y: Scalar = Field(..., description="Y-coordinate of the top-left corner.")
    width: Scalar = Field(..., description="Width of the rectangle.")
    height: Scalar = Field(..., description="Height of the rectangle.")

    def shifted(self, dx: float, dy: float) -> RectangleRoi:
        """Return a new rectangle shifted by the given amount."""
        return self.model_copy(update={"x": self.x + dx, "y": self.y + dy})

    def area(self) -> float:
        """Return the area of the rectangle."""
        return self.width * self.height

    def bbox(self) -> Rect[float]:
        """Return the bounding box of the rectangle."""
        return Rect(self.x, self.y, self.width, self.height)

    def to_mask(self, shape: tuple[int, ...]) -> NDArray[np.bool_]:
        bb = self.bbox().adjust_to_int("inner")
        arr = np.zeros(shape, dtype=bool)
        arr[..., bb.top : bb.bottom, bb.left : bb.right] = True
        return arr


class RotatedRoi2D(Roi2D):
    start: tuple[float, float] = Field(..., description="(X, Y) coordinate of the start point.")  # fmt: skip
    end: tuple[float, float] = Field(..., description="(X, Y) coordinate of the end point.")  # fmt: skip
    width: float = Field(..., description="Width of the ROI.")

    def length(self) -> float:
        start_x, start_y = self.start
        end_x, end_y = self.end
        return math.hypot(end_x - start_x, end_y - start_y)

    def shifted(self, dx: float, dy: float) -> RotatedRectangleRoi:
        start = (self.start[0] + dx, self.start[1] + dy)
        end = (self.end[0] + dx, self.end[1] + dy)
        return self.model_copy(update={"start": start, "end": end})

    def _get_vx_vy(self):
        start_x, start_y = self.start
        end_x, end_y = self.end
        length = math.hypot(end_x - start_x, end_y - start_y)
        rad = self.angle_radian()
        vx = np.array([math.cos(rad), math.sin(rad)]) * length
        vy = np.array([-math.sin(rad), math.cos(rad)]) * self.width
        return vx, vy

    def angle(self) -> float:
        """Counter-clockwise rotation."""
        return math.degrees(self.angle_radian())

    def angle_radian(self) -> float:
        # NOTE: invert y so that angle is CCW
        return math.atan2(self.start[1] - self.end[1], self.end[0] - self.start[0])


class RotatedRectangleRoi(RotatedRoi2D):
    """ROI that represents a rotated rectangle."""

    def area(self) -> float:
        return self.length() * self.width

    def bbox(self) -> Rect[float]:
        p00, p01, p11, p10 = self._get_vertices()
        xmin = min(p00[0], p01[0], p10[0], p11[0])
        xmax = max(p00[0], p01[0], p10[0], p11[0])
        ymin = min(p00[1], p01[1], p10[1], p11[1])
        ymax = max(p00[1], p01[1], p10[1], p11[1])
        return Rect(xmin, ymin, xmax - xmin, ymax - ymin)

    def _get_vertices(self):
        start_x, start_y = self.start
        end_x, end_y = self.end
        vx, vy = self._get_vx_vy()
        center = np.array([start_x + end_x, start_y + end_y]) / 2
        p00 = center - vx / 2 - vy / 2
        p01 = center - vx / 2 + vy / 2
        p10 = center + vx / 2 - vy / 2
        p11 = center + vx / 2 + vy / 2
        return p00, p01, p11, p10

    def to_mask(self, shape: tuple[int, ...]):
        vertices = np.stack(self._get_vertices(), axis=0)
        return _utils.polygon_mask(shape, vertices[:, ::-1])


class EllipseRoi(Roi2D):
    """ROI that represents an ellipse."""

    x: Scalar = Field(..., description="X-coordinate of the left boundary.")
    y: Scalar = Field(..., description="Y-coordinate of the top boundary.")
    width: Scalar = Field(..., description="Diameter along the x-axis.")
    height: Scalar = Field(..., description="Diameter along the y-axis.")

    def center(self) -> tuple[float, float]:
        return self.x + self.width / 2, self.y + self.height / 2

    def shifted(self, dx: float, dy: float) -> EllipseRoi:
        return self.model_copy(update={"x": self.x + dx, "y": self.y + dy})

    def area(self) -> float:
        return math.pi * self.width * self.height / 4

    def eccentricity(self) -> float:
        """Eccentricity of the ellipse."""
        return _utils.eccentricity(self.width / 2, self.height / 2)

    def bbox(self) -> Rect[float]:
        return Rect(self.x, self.y, self.width, self.height)

    def to_mask(self, shape: tuple[int, ...]) -> NDArray[np.bool_]:
        _yy, _xx = np.indices(shape[-2:])
        cx, cy = self.center()
        if self.height == 0 or self.width == 0:
            return np.zeros(shape, dtype=bool)
        comp_a = (_yy - cy) / self.height * 2
        comp_b = (_xx - cx) / self.width * 2
        return comp_a**2 + comp_b**2 <= 1


class RotatedEllipseRoi(RotatedRoi2D):
    """ROI that represents a rotated ellipse."""

    def area(self) -> float:
        return self.length() * self.width * math.pi / 4

    def to_mask(self, shape: tuple[int, ...]) -> NDArray[np.bool_]:
        _yy, _xx = np.indices(shape[-2:])
        start_x, start_y = self.start
        end_x, end_y = self.end
        length = math.hypot(end_x - start_x, end_y - start_y)
        cx, cy = (start_x + end_x) / 2, (start_y + end_y) / 2
        angle = self.angle_radian()
        comp_a = (_yy - cy) / length * 2
        comp_b = (_xx - cx) / self.width * 2
        comp_a, comp_b = (
            comp_a * math.cos(angle) - comp_b * math.sin(angle),
            comp_a * math.sin(angle) + comp_b * math.cos(angle),
        )
        return comp_a**2 + comp_b**2 <= 1

    def eccentricity(self) -> float:
        """Eccentricity of the ellipse."""
        return _utils.eccentricity(self.length() / 2, self.width / 2)


class PointRoi2D(Roi2D):
    """ROI that represents a single point."""

    x: float = Field(..., description="X-coordinate of the point.")
    y: float = Field(..., description="Y-coordinate of the point.")

    def shifted(self, dx: float, dy: float) -> PointRoi2D:
        return self.model_copy(update={"x": self.x + dx, "y": self.y + dy})

    def bbox(self) -> Rect[float]:
        return Rect(self.x, self.y, 0, 0)

    def to_mask(self, shape: tuple[int, ...]) -> NDArray[np.bool_]:
        arr = np.zeros(shape, dtype=bool)
        arr[..., int(round(self.y)), int(round(self.x))] = True
        return arr


class PointsRoi2D(Roi2D):
    """ROI that represents a set of points."""

    xs: Any = Field(..., description="List of x-coordinates.")
    ys: Any = Field(..., description="List of y-coordinates.")

    @field_validator("xs", "ys")
    def _validate_np_arrays(cls, v) -> NDArray[np.number]:
        out = np.asarray(v)
        if out.dtype.kind not in "if":
            raise ValueError("Must be a numerical array.")
        return out

    def model_dump_typed(self) -> dict[str, Any]:
        out = super().model_dump_typed()
        out["xs"] = self.xs.tolist()
        out["ys"] = self.ys.tolist()
        return out

    def shifted(self, dx: float, dy: float) -> PointsRoi2D:
        return self.model_copy(update={"xs": self.xs + dx, "ys": self.ys + dy})

    def bbox(self) -> Rect[float]:
        xmin, xmax = np.min(self.xs), np.max(self.xs)
        ymin, ymax = np.min(self.ys), np.max(self.ys)
        return Rect(xmin, ymin, xmax - xmin, ymax - ymin)

    def to_mask(self, shape: tuple[int, ...]) -> NDArray[np.bool_]:
        arr = np.zeros(shape, dtype=bool)
        xs = np.asarray(self.xs).round().astype(int)
        ys = np.asarray(self.ys).round().astype(int)
        arr[..., ys, xs] = True
        return arr


class CircleRoi(Roi2D):
    """ROI that represents a circle."""

    x: Scalar = Field(..., description="X-coordinate of the center.")
    y: Scalar = Field(..., description="Y-coordinate of the center.")
    radius: Scalar = Field(..., description="Radius of the circle.")

    def shifted(self, dx: float, dy: float) -> CircleRoi:
        return self.model_copy(update={"x": self.x + dx, "y": self.y + dy})

    def area(self) -> float:
        return math.pi * self.radius**2

    def circumference(self) -> float:
        return 2 * math.pi * self.radius

    def bbox(self) -> Rect[float]:
        return Rect(
            self.x - self.radius,
            self.y - self.radius,
            2 * self.radius,
            2 * self.radius,
        )

    def to_mask(self, shape: tuple[int, ...]) -> NDArray[np.bool_]:
        _yy, _xx = np.indices(shape[-2:])
        comp_a = (_yy - self.y) / self.radius
        comp_b = (_xx - self.x) / self.radius
        return comp_a**2 + comp_b**2 <= 1


class LineRoi(Roi2D):
    """A 2D line ROI."""

    start: tuple[float, float] = Field(..., description="(X, Y) coordinate of the start point.")  # fmt: skip
    end: tuple[float, float] = Field(..., description="(X, Y) coordinate of the end point.")  # fmt: skip

    @property
    def x1(self) -> float:
        return self.start[0]

    @property
    def y1(self) -> float:
        return self.start[1]

    @property
    def x2(self) -> float:
        return self.end[0]

    @property
    def y2(self) -> float:
        return self.end[1]

    def shifted(self, dx: float, dy: float) -> LineRoi:
        """Shift the line by the given amount."""
        return LineRoi(
            start=(self.x1 + dx, self.y1 + dy),
            end=(self.x2 + dx, self.y2 + dy),
        )

    def length(self) -> float:
        """Length of the line."""
        dx = self.x2 - self.x1
        dy = self.y2 - self.y1
        return math.hypot(dx, dy)

    def angle(self) -> float:
        """Angle in degrees."""
        return math.degrees(self.angle_radian())

    def angle_radian(self) -> float:
        dx = self.x2 - self.x1
        dy = self.y1 - self.y2  # NOTE: invert y so that angle is CCW
        return math.atan2(dy, dx)

    def linspace(self, num: int) -> tuple[NDArray[np.float64], NDArray[np.float64]]:
        """Return a tuple of x and y coordinates of np.linspace along the line."""
        return np.linspace(self.x1, self.x2, num), np.linspace(self.y1, self.y2, num)

    def arange(
        self, step: float = 1.0
    ) -> tuple[NDArray[np.float64], NDArray[np.float64]]:
        """Return a tuple of x and y coordinates of np.arange along the line."""
        radian = -self.angle_radian()
        num, rem = divmod(self.length(), step)
        xrem = rem * math.cos(radian)
        yrem = rem * math.sin(radian)
        return (
            np.linspace(self.x1, self.x2 - xrem, int(num) + 1),
            np.linspace(self.y1, self.y2 - yrem, int(num) + 1),
        )

    def bbox(self) -> Rect[float]:
        xmin, xmax = min(self.x1, self.x2), max(self.x1, self.x2)
        ymin, ymax = min(self.y1, self.y2), max(self.y1, self.y2)
        return Rect(xmin, ymin, xmax - xmin, ymax - ymin)

    def to_mask(self, shape: tuple[int, ...]) -> NDArray[np.bool_]:
        arr = np.zeros(shape, dtype=bool)
        xs, ys = self.linspace(int(self.length() + 1))
        xs = xs.round().astype(int)
        ys = ys.round().astype(int)
        arr[ys, xs] = True
        return arr


class SegmentedLineRoi(PointsRoi2D):
    """ROI that represents a segmented line."""

    def length(self) -> np.float64:
        return np.sum(self.lengths())

    def lengths(self) -> NDArray[np.float64]:
        return np.hypot(np.diff(self.xs), np.diff(self.ys))

    def linspace(self, num: int) -> tuple[NDArray[np.float64], NDArray[np.float64]]:
        """Return a tuple of x and y coordinates of np.linspace along the line."""
        tnots = np.cumsum(np.concatenate([[0], self.lengths()], dtype=np.float64))
        teval = np.linspace(0, tnots[-1], num)
        xi = np.interp(teval, tnots, self.xs)
        yi = np.interp(teval, tnots, self.ys)
        return xi, yi

    def arange(
        self, step: float = 1.0
    ) -> tuple[NDArray[np.float64], NDArray[np.float64]]:
        tnots = np.cumsum(np.concatenate([[0], self.lengths()], dtype=np.float64))
        length = tnots[-1]
        num, rem = divmod(length, step)
        teval = np.linspace(0, length - rem, int(num + 1))
        xi = np.interp(teval, tnots, self.xs)
        yi = np.interp(teval, tnots, self.ys)
        return xi, yi

    def to_mask(self, shape: tuple[int, ...]) -> NDArray[np.bool_]:
        arr = np.zeros(shape, dtype=bool)
        xs, ys = self.linspace(int(math.ceil(self.length())))
        xs = xs.round().astype(int)
        ys = ys.round().astype(int)
        arr[ys, xs] = True
        return arr


class PolygonRoi(SegmentedLineRoi):
    """ROI that represents a closed polygon."""

    def to_mask(self, shape: tuple[int, ...]) -> NDArray[np.bool_]:
        return _utils.polygon_mask(shape, np.column_stack((self.ys, self.xs)))

    def area(self) -> float:
        dot_xy = np.dot(self.xs, np.roll(self.ys, 1))
        dot_yx = np.dot(self.ys, np.roll(self.xs, 1))
        return np.abs(dot_xy - dot_yx) / 2


class SplineRoi(Roi2D):
    """ROI that represents a spline curve."""

    degree: int = Field(3, description="Degree of the spline curve.", ge=1)



import numpy as np
from numpy.testing import assert_equal
from pytestqt.qtbot import QtBot
from himena import MainWindow
from himena.consts import StandardType
from himena.core import create_table_model
from himena.standards.model_meta import TableMeta
from himena.testing import WidgetTester, table
from himena.types import WidgetDataModel
from himena_builtins.qt.table import QSpreadsheet
from qtpy.QtWidgets import QApplication
from qtpy.QtCore import Qt, QPoint

_Ctrl = Qt.KeyboardModifier.ControlModifier


def test_table_edit(himena_ui: MainWindow, qtbot: QtBot):
    with _get_tester(himena_ui) as tester:
        tester.update_model(value=[["a", "b"], [0, 1]])
        tester.cycle_model()
        qtbot.addWidget(tester.widget)
        tester.widget.selection_model.current_index = (2, 3)
        tester.widget.model().setData(
            tester.widget.model().index(2, 3), "a", Qt.ItemDataRole.EditRole
        )
        assert tester.widget.to_model().value[2, 3] == "a"
        tester.widget.undo()
        tester.widget.redo()
        tester.widget._make_context_menu()
        qtbot.keyClick(tester.widget, Qt.Key.Key_A, modifier=_Ctrl)
        qtbot.keyClick(tester.widget, Qt.Key.Key_C, modifier=_Ctrl)
        qtbot.keyClick(tester.widget, Qt.Key.Key_X, modifier=_Ctrl)
        qtbot.keyClick(tester.widget, Qt.Key.Key_V, modifier=_Ctrl)
        qtbot.keyClick(tester.widget, Qt.Key.Key_Delete)
        qtbot.keyClick(tester.widget, Qt.Key.Key_F, modifier=_Ctrl)
        tester.widget.resize(100, 100)
        tester.widget._insert_row_above()
        tester.widget.undo()
        tester.widget.redo()
        tester.widget._insert_row_below()
        tester.widget.undo()
        tester.widget.redo()
        tester.widget._insert_column_left()
        tester.widget.undo()
        tester.widget.redo()
        tester.widget._insert_column_right()
        tester.widget.undo()
        tester.widget.redo()
        tester.widget.selection_model.current_index = (1, 1)
        tester.widget._remove_selected_rows()
        tester.widget.undo()
        tester.widget.redo()
        tester.widget.selection_model.current_index = (1, 1)
        tester.widget._remove_selected_columns()
        tester.widget.undo()
        tester.widget.redo()
        qtbot.keyClick(tester.widget, Qt.Key.Key_Z, modifier=_Ctrl)
        qtbot.keyClick(tester.widget, Qt.Key.Key_Y, modifier=_Ctrl)
        qtbot.keyClick(tester.widget, Qt.Key.Key_E)
        tester.widget._horizontal_header._update_press_release(Qt.KeyboardModifier.ControlModifier)

        tester.widget._set_status_tip_for_text("a", ctrl_down=False)
        tester.widget._set_status_tip_for_text("b", ctrl_down=True)
        tester.widget._set_status_tip_for_text("", ctrl_down=False)
        tester.widget._set_status_tip_for_text("", ctrl_down=True)
        tester.widget._mouse_track.last_click_pos = QPoint(20, 10)
        assert tester.widget._mouse_track.is_close_to(QPoint(20, 11))
        for last_btn in [None, "left", "right", "middle"]:
            tester.widget._mouse_track.last_button = last_btn
            tester.widget._mouse_move_event(QPoint(10, 10), ctrl_down=False)
            tester.widget._mouse_move_event(QPoint(10, 20), ctrl_down=False)
            tester.widget._mouse_move_event(QPoint(10, 10), ctrl_down=True)
            tester.widget._mouse_move_event(QPoint(10, 20), ctrl_down=True)
        tester.widget._mouse_press_event(QPoint(10, 10), Qt.MouseButton.LeftButton)
        tester.widget._mouse_press_event(QPoint(10, 10), Qt.MouseButton.RightButton)
        tester.widget._mouse_press_event(QPoint(10, 10), Qt.MouseButton.MiddleButton)

def test_moving_in_table(himena_ui: MainWindow, qtbot: QtBot):
    with _get_tester(himena_ui) as tester:
        qtbot.addWidget(tester.widget)
        tester.widget.show()
        tester.update_model(value=[["a", "b"], ["c", "bc"]])
        tester.cycle_model()
        qtbot.addWidget(tester.widget)
        tester.widget.selection_model.current_index = (0, 0)
        qtbot.keyClick(tester.widget, Qt.Key.Key_Right)
        qtbot.keyClick(tester.widget, Qt.Key.Key_Left)
        qtbot.keyClick(tester.widget, Qt.Key.Key_Down)
        qtbot.keyClick(tester.widget, Qt.Key.Key_Up)
        qtbot.keyClick(tester.widget, Qt.Key.Key_Right, modifier=_Ctrl)
        qtbot.keyClick(tester.widget, Qt.Key.Key_Left, modifier=_Ctrl)
        qtbot.keyClick(tester.widget, Qt.Key.Key_Down, modifier=_Ctrl)
        qtbot.keyClick(tester.widget, Qt.Key.Key_Up, modifier=_Ctrl)
        qtbot.keyClick(tester.widget, Qt.Key.Key_Home)
        qtbot.keyClick(tester.widget, Qt.Key.Key_End)
        qtbot.keyClick(tester.widget, Qt.Key.Key_PageUp)
        qtbot.keyClick(tester.widget, Qt.Key.Key_PageDown)

def test_find_table(himena_ui: MainWindow, qtbot: QtBot):
    with _get_tester(himena_ui) as tester:
        tester.update_model(value=[["a", "b"], ["c", "bc"]])
        tester.cycle_model()
        qtbot.addWidget(tester.widget)
        qtbot.keyClick(tester.widget, Qt.Key.Key_F, modifier=_Ctrl)
        finder = tester.widget._finder_widget
        assert finder is not None
        finder._line_edit.setText("b")
        qtbot.keyClick(finder, Qt.Key.Key_Enter)
        qtbot.keyClick(finder, Qt.Key.Key_Enter, modifier=Qt.KeyboardModifier.ShiftModifier)
        finder._btn_next.click()
        finder._btn_prev.click()

def test_table_view_accepts_table_like(himena_ui: MainWindow):
    table.test_accepts_table_like(_get_tester(himena_ui))

def test_table_view_current_position(himena_ui: MainWindow):
    table.test_current_position(_get_tester(himena_ui))

def test_table_view_selections(himena_ui: MainWindow):
    table.test_selections(_get_tester(himena_ui))

def test_copy_and_paste(himena_ui: MainWindow, qtbot: QtBot):
    tester = _get_tester(himena_ui)
    qtbot.addWidget(tester.widget)
    tester.update_model(value=[["a", "b"], ["c", "bc"]])
    tester.widget.selection_model.current_index = (0, 0)
    tester.widget.selection_model.set_ranges([(slice(0, 1), slice(0, 1))])
    tester.widget._copy_as_csv()
    tester.widget._copy_as_markdown()
    tester.widget._copy_as_html()
    tester.widget._copy_as_rst()
    tester.widget._copy_to_clipboard()
    QApplication.processEvents()
    tester.widget.selection_model.current_index = (1, 1)
    tester.widget.selection_model.set_ranges([(slice(1, 2), slice(1, 2))])
    tester.widget._paste_from_clipboard()
    assert_equal(tester.widget.to_model().value, [["a", "b"], ["c", "a"]])
    tester.widget.selection_model.set_ranges([(slice(0, 2), slice(0, 3))])
    tester.widget._paste_from_clipboard()
    assert_equal(tester.widget.to_model().value, [["a", "a", "a"], ["a", "a", "a"]])
    tester.update_model(value=[["a", "b"], ["c", "bc"]])
    tester.widget.selection_model.set_ranges([(slice(0, 2), slice(0, 1))])
    tester.widget._copy_to_clipboard()
    QApplication.processEvents()
    tester.widget.selection_model.set_ranges([(slice(0, 2), slice(1, 2))])
    tester.widget._paste_from_clipboard()
    assert_equal(tester.widget.to_model().value, [["a", "a"], ["c", "c"]])

def _get_tester(himena_ui: MainWindow):
    return WidgetTester(QSpreadsheet(himena_ui))

def test_commands(himena_ui: MainWindow):
    model = WidgetDataModel(
        value=[["a", "b", "c"], ["d", "e", "f"]],
        type=StandardType.TABLE,
        metadata=TableMeta(selections=[], separator="\t")
    )
    himena_ui.add_data_model(model)
    himena_ui.exec_action("builtins:table:copy-as-csv")
    model = WidgetDataModel(
        value=[["a", "b", "c"], ["d", "e", "f"]],
        type=StandardType.TABLE,
        metadata=TableMeta(selections=[((0, 1), (1, 2))], separator=",")
    )
    himena_ui.add_data_model(model)
    himena_ui.exec_action("builtins:table:copy-as-csv")
    himena_ui.exec_action("builtins:table:copy-as-markdown")
    himena_ui.exec_action("builtins:table:copy-as-html")
    himena_ui.exec_action("builtins:table:copy-as-rst")
    himena_ui.exec_action("builtins:table:crop")
    himena_ui.exec_action("builtins:table:change-separator", with_params={"separator": "\t"})
    himena_ui.exec_action(
        "builtins:table:insert-incrementing-numbers",
        with_params={"selection": ((0, 1), (1, 4)), "start": 1, "step": 2}
    )
    assert_equal(himena_ui.current_model.value[0, 1:4], ["1", "3", "5"])
    himena_ui.exec_action(
        "builtins:table:insert-incrementing-numbers",
        with_params={"selection": ((0, 10), (1, 2)), "start": 1, "step": 1}
    )
    assert_equal(himena_ui.current_model.value[0:10, 1], [str(i) for i in range(1, 11)])

    model = WidgetDataModel(
        value=[["a", "3", "-2"], ["d", "4", "1.1"]],
        type=StandardType.TABLE,
        metadata=TableMeta(selections=[], separator="\t")
    )
    himena_ui.add_data_model(model)
    widget = himena_ui.current_window.widget
    assert isinstance(widget, QSpreadsheet)
    widget._selection_model.set_ranges([(slice(1, 3), slice(1, 2))])
    widget._measure()

def test_large_data(himena_ui: MainWindow, qtbot: QtBot):
    # initialize with a large data
    ss = QSpreadsheet(himena_ui)
    qtbot.addWidget(ss)
    ss.update_model(
        WidgetDataModel(
            value=[["a"] * 100] * 1000,
            type=StandardType.TABLE,
        )
    )
    assert ss.model().rowCount() == 1001
    assert ss.model().columnCount() == 101

    # paste a large data
    ss = QSpreadsheet(himena_ui)
    qtbot.addWidget(ss)
    ss.update_model(WidgetDataModel(value=[["a"]], type=StandardType.TABLE))
    ss.setCurrentIndex(ss.model().index(0, 0))
    row_count_old = ss.model().rowCount()
    col_count_old = ss.model().columnCount()
    row = "\t".join(str(i) for i in range(194))
    data = "\n".join([row for _ in range(183)])
    QApplication.clipboard().setText(data)
    ss._selection_model.set_ranges([(slice(0, 1), slice(0, 1))])
    ss._paste_from_clipboard()
    assert ss.model().rowCount() == 184
    assert ss.model().columnCount() == 195
    ss.undo()
    assert ss.model().rowCount() == row_count_old
    assert ss.model().columnCount() == col_count_old
    ss.redo()
    assert ss.model().rowCount() == 184
    assert ss.model().columnCount() == 195

def test_table_deletion_at_edges(himena_ui: MainWindow, qtbot: QtBot):
    # test deleting at the edges of the table
    ss = QSpreadsheet(himena_ui)
    qtbot.addWidget(ss)
    # "a" "b" ""
    # "c" [d] "e" <- delete this cell
    ss.update_model(WidgetDataModel(value=[["a", "b", ""], ["c", "d", "e"]], type=StandardType.TABLE))
    ss._selection_model.set_ranges([(slice(1, 2), slice(1, 2))])
    ss._delete_selection()
    assert_equal(ss.to_model().value, [["a", "b", ""], ["c", "", "e"]])

    # "a" "b" ""
    # "c" "d" [e] <- delete this cell
    ss.update_model(WidgetDataModel(value=[["a", "b", ""], ["c", "d", "e"]], type=StandardType.TABLE))
    ss._selection_model.set_ranges([(slice(1, 2), slice(2, 3))])
    ss._delete_selection()
    assert_equal(ss.to_model().value, [["a", "b"], ["c", "d"]])

    # "a" "b" ""
    # "c" "d" "e"
    # ""  [f] "" <- delete this cell
    ss.update_model(
        WidgetDataModel(
            value=[["a", "b", ""], ["c", "d", "e"], ["", "f", ""]],
            type=StandardType.TABLE,
        )
    )
    ss._selection_model.set_ranges([(slice(2, 3), slice(1, 2))])
    ss._delete_selection()
    assert_equal(ss.to_model().value, [["a", "b", ""], ["c", "d", "e"]])

    # "a" "b" ""
    # "" "" ""
    # "" "" [c] <- delete this cell
    ss.update_model(
        WidgetDataModel(
            value=[["a", "b", ""], ["", "", ""], ["", "", "c"]],
            type=StandardType.TABLE,
        )
    )
    ss._selection_model.set_ranges([(slice(2, 3), slice(2, 3))])
    ss._delete_selection()
    assert_equal(ss.to_model().value, [["a", "b"]])

def test_copy_on_write(himena_ui: MainWindow, qtbot: QtBot):
    # test copy on write
    ss = QSpreadsheet(himena_ui)
    qtbot.addWidget(ss)
    with WidgetTester(ss) as tester:
        array_orig = np.array([["a", "b"], ["c", "d"]], dtype=np.dtypes.StringDType())
        tester.update_model(value=array_orig, type=StandardType.TABLE)
        ss._selection_model.set_ranges([(slice(0, 1), slice(1, 2))])
        ss._delete_selection()
        assert_equal(ss.to_model().value, [["a", ""], ["c", "d"]])
        assert_equal(array_orig, [["a", "b"], ["c", "d"]])

        ss_other = QSpreadsheet(himena_ui)
        ss_other.update_model(ss.to_model())
        ss._selection_model.set_ranges([(slice(0, 1), slice(0, 1))])
        ss._delete_selection()
        assert_equal(ss.to_model().value, [["", ""], ["c", "d"]])
        assert_equal(ss_other.to_model().value, [["a", ""], ["c", "d"]])

def test_header_view(himena_ui: MainWindow, qtbot: QtBot):
    from himena_builtins.qt.widgets._table_components._header import (
        QHorizontalHeaderView,
        QVerticalHeaderView,
    )
    header = QHorizontalHeaderView(QSpreadsheet(himena_ui))
    qtbot.addWidget(header)
    header._on_section_clicked(0)
    header._on_section_pressed(0)
    header._on_section_entered(1)
    header.visualRectAtIndex(0)

    header = QVerticalHeaderView(QSpreadsheet(himena_ui))
    qtbot.addWidget(header)
    header._on_section_clicked(0)
    header._on_section_pressed(0)
    header._on_section_entered(1)
    header.visualRectAtIndex(0)

def test_table_view_mouse_interaction(himena_ui: MainWindow, qtbot: QtBot):

    ss = QSpreadsheet(himena_ui)
    qtbot.addWidget(ss)
    ss.show()
    ss.update_model(create_table_model(value=["a", "b"]))
    ss.update_model(create_table_model(value=np.zeros((10, 10))))

    qtbot.mouseClick(ss, Qt.MouseButton.LeftButton, pos=QPoint(5, 5))
    qtbot.mousePress(ss, Qt.MouseButton.LeftButton, pos=QPoint(5, 5))
    qtbot.mouseMove(ss, pos=QPoint(35, 55))
    qtbot.mouseRelease(ss, Qt.MouseButton.LeftButton, pos=QPoint(35, 55))

    qtbot.mousePress(ss, Qt.MouseButton.RightButton, pos=QPoint(5, 5))
    qtbot.mouseMove(ss, pos=QPoint(35, 55))
    qtbot.mouseRelease(ss, Qt.MouseButton.RightButton, pos=QPoint(35, 55))



from __future__ import annotations

from matplotlib.figure import Figure
from matplotlib.backends import backend_qtagg
import matplotlib.pyplot as plt
from qtpy import QtWidgets as QtW, QtGui, QtCore

from himena.plugins import validate_protocol
from himena.types import DropResult, Size, WidgetDataModel
from himena.consts import StandardType
from himena.style import Theme
from himena.standards import plotting as hplt
from himena.standards.model_meta import DimAxis
from himena_builtins.qt.plot._conversion import (
    convert_plot_layout,
    update_model_axis_by_mpl,
    convert_plot_model,
)
from himena_builtins.qt.plot._config import MatplotlibCanvasConfigs
from himena_builtins.qt.widgets._dim_sliders import QDimsSlider


class QMatplotlibCanvasBase(QtW.QWidget):
    def __init__(self):
        super().__init__()
        layout = QtW.QVBoxLayout(self)
        layout.setContentsMargins(0, 0, 0, 0)
        self._canvas: FigureCanvasQTAgg | None = None
        self._toolbar: backend_qtagg.NavigationToolbar2QT | None = None
        self._plot_models: hplt.BaseLayoutModel | None = None
        self._modified = False
        self._cfg = MatplotlibCanvasConfigs()
        self._last_mouse_pos = QtCore.QPoint()

    @property
    def figure(self) -> Figure:
        return self._canvas.figure

    @validate_protocol
    def control_widget(self) -> QtW.QWidget:
        return self._toolbar

    @validate_protocol
    def size_hint(self) -> tuple[int, int]:
        return 300, 240

    @validate_protocol
    def widget_resized_callback(self, size_old, size_new: tuple[int, int]):
        if size_new[0] > 40 and size_new[1] > 40:
            self._canvas.figure.tight_layout()

    def _prep_toolbar(self, toolbar_class=backend_qtagg.NavigationToolbar2QT):
        if self._toolbar is not None:
            return self._toolbar
        toolbar = toolbar_class(self._canvas, self)
        spacer = QtW.QWidget()
        toolbar.insertWidget(toolbar.actions()[0], spacer)
        return toolbar

    @validate_protocol
    def theme_changed_callback(self, theme: Theme):
        if self._toolbar is None:
            return

        icon_color = (
            QtGui.QColor(0, 0, 0)
            if theme.is_light_background()
            else QtGui.QColor(255, 255, 255)
        )
        for toolbtn in self._toolbar.findChildren(QtW.QToolButton):
            assert isinstance(toolbtn, QtW.QToolButton)
            icon = toolbtn.icon()
            pixmap = icon.pixmap(100, 100)
            mask = pixmap.mask()
            pixmap.fill(icon_color)
            pixmap.setMask(mask)
            icon_new = QtGui.QIcon(pixmap)
            toolbtn.setIcon(icon_new)
            # Setting icon to the action as well; otherwise checking/unchecking will
            # revert the icon to the original color
            toolbtn.actions()[0].setIcon(icon_new)

    def _init_canvas_and_toolbar(self):
        self._canvas = FigureCanvasQTAgg()
        self._canvas.contextmenu_requested.connect(self._show_context_menu)
        self.layout().addWidget(self._canvas)
        self._toolbar = self._prep_toolbar(QNavigationToolBar)

    def _make_context_menu(self) -> QtW.QMenu:
        menu = QtW.QMenu(self)
        menu.setTitle("Matplotlib Canvas")
        action = menu.addAction("Reset original view")
        action.triggered.connect(self._toolbar.home)
        action = menu.addAction("Back to previous view")
        action.triggered.connect(self._toolbar.back)
        action = menu.addAction("Forward to next view")
        action.triggered.connect(self._toolbar.forward)
        menu.addSeparator()
        action = menu.addAction("Copy to clipboard")
        action.triggered.connect(self._copy_canvas)
        action = menu.addAction("Save figure")
        action.triggered.connect(self._toolbar.save_figure)
        return menu

    def _show_context_menu(self, pos: QtCore.QPoint):
        self._make_context_menu().exec(pos)

    def _copy_canvas(self):
        clipboard = QtGui.QGuiApplication.clipboard()
        image = self._canvas.grab()
        clipboard.setPixmap(image, QtGui.QClipboard.Mode.Clipboard)


class QMatplotlibCanvas(QMatplotlibCanvasBase):
    """A widget that displays a Matplotlib figure itself."""

    __himena_widget_id__ = "builtins:QMatplotlibCanvasBase"
    __himena_display_name__ = "Matplotlib Canvas"

    @validate_protocol
    def update_model(self, model: WidgetDataModel):
        was_none = self._canvas is None
        if was_none:
            self._canvas = FigureCanvasQTAgg(model.value)
            self._canvas.contextmenu_requested.connect(self._show_context_menu)
            self.layout().addWidget(self._canvas)
            self._toolbar = self._prep_toolbar()
        if isinstance(model.value, Figure):
            if not was_none:
                raise ValueError("Figure is already set")
        else:
            raise ValueError(f"Unsupported model: {model.value}")
        if was_none:
            self._toolbar.pan()
            self._canvas.figure.tight_layout()

    @validate_protocol
    def to_model(self) -> WidgetDataModel:
        return WidgetDataModel(
            value=self.figure,
            type=self.model_type(),
        )

    @validate_protocol
    def model_type(self) -> str:
        return StandardType.MPL_FIGURE


class QModelMatplotlibCanvas(QMatplotlibCanvasBase):
    """A widget that displays himena standard plot models in a Matplotlib figure.

    The internal data structure is follows the himena standard.

    ## Basic Usage

    - Mouse interactivity can be controlled in the toolbar.
    - Double-click the canvas to adjust the layout.
    - This widget accepts dropping another plot model. The dropped model will be merged.

    """

    __himena_widget_id__ = "builtins:QModelMatplotlibCanvas"
    __himena_display_name__ = "Built-in Plot Canvas"

    @validate_protocol
    def update_model(self, model: WidgetDataModel):
        was_none = self._canvas is None
        _assert_plot_model(model.value)
        if was_none:
            self._init_canvas_and_toolbar()
        with plt.style.context(self._cfg.to_dict()):
            self._plot_models = convert_plot_layout(model.value, self.figure)
        self._canvas.draw()
        if was_none and not isinstance(self._plot_models, hplt.SingleAxes3D):
            self._toolbar.pan()

    @validate_protocol
    def to_model(self) -> WidgetDataModel:
        value = self._plot_models.model_copy()
        # TODO: update the model with the current canvas state as much as possible
        if isinstance(value, hplt.SingleAxes):
            model_axes_ref = [value.axes]
        elif isinstance(value, hplt.layout.Layout1D):
            model_axes_ref = value.axes
        elif isinstance(value, hplt.layout.Grid):
            model_axes_ref = sum(value.axes, [])
        elif isinstance(value, hplt.SingleAxes3D):
            model_axes_ref = [value.axes]
        else:
            model_axes_ref = []  # Not implemented
        mpl_axes_ref = self.figure.axes
        for model_axes, mpl_axes in zip(model_axes_ref, mpl_axes_ref):
            update_model_axis_by_mpl(model_axes, mpl_axes)
        return WidgetDataModel(
            value=value,
            type=self.model_type(),
            extension_default=".plot.json",
        )

    @validate_protocol
    def model_type(self) -> str:
        return StandardType.PLOT

    @validate_protocol
    def update_configs(self, cfg: MatplotlibCanvasConfigs):
        self._cfg = cfg

    @validate_protocol
    def dropped_callback(self, model: WidgetDataModel):
        if not (
            isinstance(model.value, hplt.BaseLayoutModel)
            and isinstance(self._plot_models, hplt.BaseLayoutModel)
        ):
            raise ValueError(
                f"Both models must be BaseLayoutModel, got {model.value!r} and "
                f"{self._plot_models!r}"
            )
        return DropResult(
            command_id="builtins:plot:concatenate-with", with_params={"others": [model]}
        )

    @validate_protocol
    def allowed_drop_types(self) -> list[str]:
        return [StandardType.PLOT]

    @validate_protocol
    def size_hint(self) -> tuple[int, int]:
        return 300, 240

    @validate_protocol
    def widget_resized_callback(self, size_old: Size, size_new: Size):
        if size_new.width > 40 and size_new.width > 40:
            self._canvas.figure.tight_layout()

    @validate_protocol
    def widget_added_callback(self):
        self._canvas.figure.tight_layout()

    @validate_protocol
    def is_modified(self) -> bool:
        return self._modified

    @validate_protocol
    def set_modified(self, value: bool):
        self._modified = value


class QModelMatplotlibCanvasStack(QMatplotlibCanvasBase):
    __himena_widget_id__ = "builtins:QModelMatplotlibCanvasStack"
    __himena_display_name__ = "Built-in Plot Canvas Stack"
    _plot_models: hplt.SingleStackedAxes | None

    def __init__(self):
        super().__init__()
        self._dims_slider = QDimsSlider()
        self._dims_slider.valueChanged.connect(self._slider_changed)

    @validate_protocol
    def update_model(self, model: WidgetDataModel):
        was_none = self._canvas is None
        if not isinstance(model.value, hplt.SingleStackedAxes):
            raise ValueError(
                f"Expected a SingleStackedAxes model, got {type(model.value)}"
            )
        if was_none:
            self._init_canvas_and_toolbar()
            self.layout().addWidget(self._dims_slider)

        self._dims_slider.set_dimensions(
            model.value.shape + (0, 0),
            axes=model.value.axes.multi_dims + [DimAxis(name=""), DimAxis(name="")],
        )
        with plt.style.context(self._cfg.to_dict()):
            self._plot_models = convert_plot_layout(model.value, self.figure)
        self._slider_changed(self._dims_slider.value(), auto_scale=False)
        if was_none:
            self._toolbar.pan()

    @validate_protocol
    def to_model(self) -> WidgetDataModel:
        value = self._plot_models.model_copy()
        if not isinstance(value, hplt.SingleStackedAxes):
            raise ValueError("The model is not a SingleAxesStack")
        mpl_axes_ref = self.figure.axes
        model_axes_ref = [value.axes]
        for model_axes, mpl_axes in zip(model_axes_ref, mpl_axes_ref):
            update_model_axis_by_mpl(model_axes, mpl_axes)
        return WidgetDataModel(
            value=value,
            type=self.model_type(),
            extension_default=".plot.json",
        )

    @validate_protocol
    def model_type(self) -> str:
        return StandardType.PLOT_STACK

    @validate_protocol
    def update_configs(self, cfg: MatplotlibCanvasConfigs):
        self._cfg = cfg

    def _slider_changed(self, value: int, *, auto_scale: bool = True):
        if self._plot_models is None:
            return
        if not isinstance(self._plot_models, hplt.SingleStackedAxes):
            raise ValueError("The model is not a SingleAxesStack")
        axes_component = self._plot_models.axes[value]
        ax_mpl = self.figure.axes[0]
        xlim = ax_mpl.get_xlim()
        ylim = ax_mpl.get_ylim()
        ax_mpl.clear()
        for model in axes_component.models:
            convert_plot_model(model, ax_mpl)
        if auto_scale:
            ax_mpl.set_xlim(xlim)
            ax_mpl.set_ylim(ylim)
        self._canvas.draw()


def _assert_plot_model(val):
    """Check if the value is a plot model."""
    if not isinstance(val, hplt.BaseLayoutModel):
        raise TypeError(f"Expected a plot model, got {type(val)}")


# remove some of the tool buttons
class QNavigationToolBar(backend_qtagg.NavigationToolbar2QT):
    toolitems = (
        ("Home", "Reset original view", "home", "home"),
        ("Back", "Back to previous view", "back", "back"),
        ("Forward", "Forward to next view", "forward", "forward"),
        (None, None, None, None),
        (
            "Pan",
            "Left button pans, Right button zooms\nx/y fixes axis, CTRL fixes aspect",
            "move",
            "pan",
        ),
        ("Zoom", "Zoom to rectangle\nx/y fixes axis", "zoom_to_rect", "zoom"),
        (None, None, None, None),
        ("Save", "Save the figure", "filesave", "save_figure"),
    )


class FigureCanvasQTAgg(backend_qtagg.FigureCanvasQTAgg):
    contextmenu_requested = QtCore.Signal(QtCore.QPoint)
    _last_mouse_pos = QtCore.QPoint()

    def mouseDoubleClickEvent(self, event):
        self.figure.tight_layout()
        self.draw()
        return super().mouseDoubleClickEvent(event)

    def mousePressEvent(self, a0):
        self._last_mouse_pos = a0.pos()
        return super().mousePressEvent(a0)

    def mouseReleaseEvent(self, a0: QtGui.QMouseEvent):
        if (self._last_mouse_pos - a0.pos()).manhattanLength() < 8:
            if a0.button() == QtCore.Qt.MouseButton.RightButton:
                self.contextmenu_requested.emit(self.mapToGlobal(a0.pos()))
        self._last_mouse_pos = QtCore.QPoint()
        return super().mouseReleaseEvent(a0)


FigureCanvas = FigureCanvasQTAgg


# The plt.show function will be overwriten to this.
# Modified from matplotlib_inline (BSD 3-Clause "New" or "Revised" License)
# https://github.com/ipython/matplotlib-inline
def show(close=True, block=None):
    from matplotlib._pylab_helpers import Gcf
    from himena.widgets import current_instance

    ui = current_instance()

    try:
        for figure_manager in Gcf.get_all_fig_managers():
            ui.add_object(
                figure_manager.canvas.figure,
                type=StandardType.MPL_FIGURE,
                title="Plot",
            )
    finally:
        show._called = True
        if close and Gcf.get_all_fig_managers():
            plt.close("all")


show._called = False



from app_model.types import (
    KeyBindingRule,
    KeyCode,
    KeyMod,
)
from himena.consts import MenuId
from himena.widgets import MainWindow
from himena._app_model._context import AppContext as _ctx
from himena._app_model.actions._registry import ACTIONS

GO_TO_GROUP = "10_go-to"
JUMP_TO_GROUP = "11_jump-to"


# Jump to the nth window
def make_func(n: int):
    def jump_to_nth_window(ui: MainWindow) -> None:
        if (area := ui.tabs.current()) and len(area) > n:
            area.current_index = n

    jump_to_nth_window.__name__ = f"jump_to_window_{n}"
    jump_to_nth_window.__doc__ = f"Jump to the {n}-th window in the current tab."
    jump_to_nth_window.__qualname__ = f"jump_to_window_{n}"
    jump_to_nth_window.__module__ = make_func.__module__
    return jump_to_nth_window


for n in range(10):
    th: str = "st" if n == 1 else "nd" if n == 2 else "rd" if n == 3 else "th"
    keycode = getattr(KeyCode, f"Digit{n}")
    ACTIONS.append_from_fn(
        id=f"jump-to-window-{n}",
        title=f"Jump To {n}{th} Window",
        enablement=_ctx.num_sub_windows > n,
        menus=[{"id": MenuId.GO, "group": JUMP_TO_GROUP}],
        keybindings=[{"primary": KeyMod.Alt | keycode}],
    )(make_func(n))


@ACTIONS.append_from_fn(
    id="go-to-previous-window",
    title="Go To Previous Window",
    menus=[MenuId.GO],
    enablement=(_ctx.num_tabs > 0) & (_ctx.num_sub_windows > 1),
    keybindings=[KeyBindingRule(primary=KeyMod.CtrlCmd | KeyCode.Home)],
    recording=False,
)
def go_to_previous_window(ui: MainWindow) -> None:
    """Go to the previous window."""
    if (area := ui.tabs.current()) and len(area) > 1:
        area.current_index = (area.current_index - 1) % len(area)


@ACTIONS.append_from_fn(
    id="go-to-next-window",
    title="Go To Next Window",
    menus=[MenuId.GO],
    enablement=(_ctx.num_tabs > 0) & (_ctx.num_sub_windows > 1),
    keybindings=[KeyBindingRule(primary=KeyMod.CtrlCmd | KeyCode.End)],
    recording=False,
)
def go_to_next_window(ui: MainWindow) -> None:
    """Go to the next window."""
    if (area := ui.tabs.current()) and len(area) > 1:
        area.current_index = (area.current_index + 1) % len(area)


@ACTIONS.append_from_fn(
    id="go-to-window",
    title="Go To Window ...",
    menus=[MenuId.GO],
    enablement=_ctx.num_tabs > 0,
    keybindings=[KeyBindingRule(primary=KeyMod.CtrlCmd | KeyCode.KeyG)],
    recording=False,
)
def go_to_window(ui: MainWindow) -> None:
    """Go to an existing window."""
    ui._backend_main_window._show_command_palette("goto")



from __future__ import annotations

import re
import sys
from typing import Callable, Generator, TYPE_CHECKING
import weakref

import numpy as np
from qtpy import QtWidgets as QtW, QtGui, QtCore
from psygnal import EmitLoopError

from himena.consts import MonospaceFontFamily
from himena.utils.misc import ansi2html

if TYPE_CHECKING:
    from types import TracebackType
    from warnings import WarningMessage
    from himena.qt._qmain_window import QMainWindow

    ExcInfo = tuple[type[BaseException], BaseException, TracebackType]


class QtTracebackDialog(QtW.QDialog):
    """A dialog box that shows Python traceback."""

    def __init__(self, parent):
        super().__init__(parent)
        self.setWindowTitle("Traceback")
        layout = QtW.QVBoxLayout()
        self.setLayout(layout)

        # prepare text edit
        self._text = QtW.QTextEdit(self)
        self._text.setReadOnly(True)
        self._text.setFontFamily(MonospaceFontFamily)
        self._text.setLineWrapMode(QtW.QTextEdit.LineWrapMode.NoWrap)
        layout.addWidget(self._text)

        self.resize(600, 400)

    def setText(self, text: str):
        """Always set text as a HTML text."""
        self._text.setHtml(text)

    def keyPressEvent(self, a0: QtGui.QKeyEvent) -> None:
        if (
            a0.modifiers() == QtCore.Qt.KeyboardModifier.ControlModifier
            and a0.key() in (QtCore.Qt.Key.Key_W, QtCore.Qt.Key.Key_Q)
        ):
            self.close()
        else:
            return super().keyPressEvent(a0)


class QtErrorMessageBox(QtW.QWidget):
    """An message box widget for displaying Python exception."""

    def __init__(
        self,
        text: str,
        exc: Exception | None,
        parent: QMainWindow,
    ):
        self._main_window = weakref.ref(parent)
        super().__init__()

        self._exc = exc
        self.text_edit = QtW.QPlainTextEdit(self)
        self.text_edit.setReadOnly(True)
        self.text_edit.setWordWrapMode(QtGui.QTextOption.WrapMode.WordWrap)
        self.text_edit.setVerticalScrollBarPolicy(
            QtCore.Qt.ScrollBarPolicy.ScrollBarAsNeeded
        )
        self.text_edit.setHorizontalScrollBarPolicy(
            QtCore.Qt.ScrollBarPolicy.ScrollBarAlwaysOff
        )
        self.text_edit.setPlainText(text)

        layout = QtW.QVBoxLayout(self)
        layout.addWidget(self.text_edit, stretch=100)

    def _traceback_button_clicked(self, *, runtime: bool = True):
        tb = self._get_traceback()
        dlg = QtTracebackDialog(self)
        dlg.setText(tb)
        focus = QtW.QApplication.focusWidget()
        if runtime:
            dlg.exec()
        if focus:
            focus.setFocus()

    def _enter_debugger_button_clicked(self):
        import pdb

        if hasattr(QtCore, "pyqtRemoveInputHook"):
            QtCore.pyqtRemoveInputHook()
        try:
            pdb.post_mortem(self._exc.__traceback__)
        finally:
            if hasattr(QtCore, "pyqtRestoreInputHook"):
                QtCore.pyqtRestoreInputHook()

    @classmethod
    def from_exc(cls, e: Exception, parent: QMainWindow):
        """Construct message box from a exception."""
        # unwrap EmitLoopError
        while isinstance(e, EmitLoopError):
            e = e.__cause__

        if len(e.args) > 0:
            text = f"{type(e).__name__}: {e}"
        else:
            text = ""
        self = cls(text, e, parent)

        # prepare buttons specific to Exception
        footer = QtW.QWidget()
        layout = QtW.QHBoxLayout(footer)
        layout.setContentsMargins(0, 0, 0, 0)

        traceback_button = QtW.QPushButton("Trackback", self)
        enter_debugger_button = QtW.QPushButton("Debug", self)
        traceback_button.setFixedHeight(20)
        enter_debugger_button.setFixedHeight(20)

        traceback_button.clicked.connect(self._traceback_button_clicked)
        enter_debugger_button.clicked.connect(self._enter_debugger_button_clicked)

        layout.addWidget(traceback_button)
        layout.addWidget(enter_debugger_button)
        layout.setAlignment(QtCore.Qt.AlignmentFlag.AlignRight)
        self.layout().addWidget(footer)
        return self

    @classmethod
    def from_warning(cls, w: WarningMessage, parent: QMainWindow):
        if isinstance(w.message, Warning):
            text = str(w.message)
            exc = w.message
        else:
            text = w.message
            exc = None
        self = cls(text, exc, parent)
        return self

    def _exc_info(self) -> ExcInfo:
        return (
            self._exc.__class__,
            self._exc,
            self._exc.__traceback__,
        )

    def _get_traceback(self) -> str:
        background = (
            self._main_window().palette().color(QtGui.QPalette.ColorRole.Window)
        )
        is_dark = background.lightnessF() < 0.5
        if self._exc is None:
            import traceback

            tb = traceback.format_exc()
        else:
            tb = get_tb_formatter()(self._exc_info(), as_html=True, is_dark=is_dark)
        return tb


# Following functions are mostly copied from napari (BSD 3-Clause).
# See https://github.com/napari/napari/blob/main/napari/utils/_tracebacks.py


def get_tb_formatter() -> Callable[[ExcInfo, bool, bool], str]:
    """Return a formatter callable that uses IPython VerboseTB if available.

    Imports IPython lazily if available to take advantage of ultratb.VerboseTB.
    If unavailable, cgitb is used instead, but this function overrides a lot of
    the hardcoded citgb styles and adds error chaining (for exceptions that
    result from other exceptions).

    Returns
    -------
    callable
        A function that accepts a 3-tuple and a boolean ``(exc_info, as_html)``
        and returns a formatted traceback string. The ``exc_info`` tuple is of
        the ``(type, value, traceback)`` format returned by sys.exc_info().
        The ``as_html`` determines whether the traceback is formatted in html
        or plain text.
    """
    try:
        import IPython.core.ultratb  # noqa: F401

        format_exc_info = format_exc_info_ipython

    except ModuleNotFoundError:  # pragma: no cover
        if sys.version_info < (3, 11):
            format_exc_info = format_exc_info_py310
        else:
            format_exc_info = format_exc_info_py311
    return format_exc_info


def _filter_traceback(tb: TracebackType) -> TracebackType:
    tb_orig = tb
    while tb:
        if tb.tb_next and tb.tb_next.tb_frame.f_locals.get("__tracebackhide__", False):
            tb.tb_next = None
            break
        tb = tb.tb_next
    return tb_orig


def format_exc_info_ipython(info: ExcInfo, as_html: bool, is_dark: bool = False) -> str:
    import IPython.core.ultratb

    typ, exc, tb = info
    color = "Neutral"
    tb = _filter_traceback(tb)

    # avoid verbose printing of the array data
    with np.printoptions(precision=5, threshold=10, edgeitems=2):
        if IPython.version_info >= (9, 0):
            vbtb = IPython.core.ultratb.VerboseTB(theme_name=color.lower())
        else:
            vbtb = IPython.core.ultratb.VerboseTB(color_scheme=color)
        if as_html:
            ansi_string = (
                vbtb.text(typ, exc, tb)
                .replace(" ", "&nbsp;")
                .replace("<", "&lt;")
                .replace(">", "&gt;")
                .replace("\n", "<br>")
            )
            _html = "".join(ansi2html(ansi_string, is_dark=is_dark))
            _html = (
                f"<span style='font-family: monaco,{MonospaceFontFamily},"
                "monospace;'>" + _html + "</span>"
            )
            tb_text = _html
        else:
            tb_text = vbtb.text(*info)
    return tb_text


# cgitb does not support error chaining...
# see https://peps.python.org/pep-3134/#enhanced-reporting
# this is a workaround
def cgitb_chain(exc: Exception) -> Generator[str, None, None]:
    """Recurse through exception stack and chain cgitb_html calls."""
    if exc.__cause__:
        yield from cgitb_chain(exc.__cause__)
        yield (
            '<br><br><font color="#51B432">The above exception was '
            "the direct cause of the following exception:</font><br>"
        )
    elif exc.__context__:
        yield from cgitb_chain(exc.__context__)
        yield (
            '<br><br><font color="#51B432">During handling of the '
            "above exception, another exception occurred:</font><br>"
        )
    yield cgitb_html(exc)


def cgitb_html(exc: Exception) -> str:
    """Format exception with cgitb.html."""
    import cgitb

    info = (type(exc), exc, exc.__traceback__)
    return cgitb.html(info)


def format_exc_info_py310(info: ExcInfo, as_html: bool, is_dark=False) -> str:
    import traceback

    # avoid verbose printing of the array data
    with np.printoptions(precision=5, threshold=10, edgeitems=2):
        if as_html:
            html = "\n".join(cgitb_chain(info[1]))
            # cgitb has a lot of hardcoded colors that don't work for us
            # remove bgcolor, and let theme handle it
            html = re.sub('bgcolor="#.*"', "", html)
            # remove superfluous whitespace
            html = html.replace("<br>\n", "\n")
            # but retain it around the <small> bits
            html = re.sub(r"(<tr><td><small.*</tr>)", "<br>\\1<br>", html)
            # weird 2-part syntax is a workaround for hard-to-grep text.
            html = html.replace(
                "<p>A problem occurred in a Python script.  Here is the sequence of",
                "",
            )
            html = html.replace(
                "function calls leading up to the error, "
                "in the order they occurred.</p>",
                "<br>",
            )
            # remove hardcoded fonts
            html = html.replace('face="helvetica, arial"', "")
            html = (
                "<span style='font-family: monaco,courier,monospace;'>"
                + html
                + "</span>"
            )
            tb_text = html
        else:
            # if we don't need HTML, just use traceback
            tb_text = "".join(traceback.format_exception(*info))
    return tb_text


def format_exc_info_py311(info: ExcInfo, as_html: bool, is_dark=False) -> str:
    import traceback

    # avoid verbose printing of the array data
    with np.printoptions(precision=5, threshold=10, edgeitems=2):
        tb_text = "".join(traceback.format_exception(*info))
        if as_html:
            tb_text = "<pre>" + tb_text + "</pre>"
    return tb_text



from __future__ import annotations

import os
from himena.plugins import register_widget_class
from himena.consts import StandardType
from himena.utils.misc import lru_cache
from himena_builtins.qt.plot._config import MatplotlibCanvasConfigs

BACKEND_HIMENA = "module://himena_builtins.qt.plot._canvas"


@lru_cache(maxsize=1)
def _is_matplotlib_available() -> bool:
    import importlib.metadata

    try:
        importlib.metadata.distribution("matplotlib")
    except importlib.metadata.PackageNotFoundError:
        return False
    return True


def register_mpl_widget():
    # Update the matplotlib default backend
    os.environ["MPLBACKEND"] = BACKEND_HIMENA
    if not _is_matplotlib_available():
        return

    register_widget_class(
        StandardType.PLOT,
        model_matplotlib_canvas,
        priority=0,
        plugin_configs=MatplotlibCanvasConfigs(),
    )
    register_widget_class(
        StandardType.PLOT_STACK,
        model_mpl_canvas_stack,
        priority=0,
        plugin_configs=MatplotlibCanvasConfigs(),
    )
    register_widget_class(StandardType.MPL_FIGURE, matplotlib_canvas, priority=0)


def model_matplotlib_canvas():
    from himena_builtins.qt.plot._canvas import QModelMatplotlibCanvas

    return QModelMatplotlibCanvas()


model_matplotlib_canvas.__himena_widget_id__ = "builtins:QModelMatplotlibCanvas"
model_matplotlib_canvas.__himena_display_name__ = "Built-in Plot Canvas"


def model_mpl_canvas_stack():
    from himena_builtins.qt.plot._canvas import QModelMatplotlibCanvasStack

    return QModelMatplotlibCanvasStack()


model_mpl_canvas_stack.__himena_widget_id__ = "builtins:QModelMatplotlibCanvasStack"
model_mpl_canvas_stack.__himena_display_name__ = "Built-in Plot Canvas Stack"


def matplotlib_canvas():
    from himena_builtins.qt.plot._canvas import QMatplotlibCanvas

    return QMatplotlibCanvas()


matplotlib_canvas.__himena_widget_id__ = "builtins:QMatplotlibCanvas"
matplotlib_canvas.__himena_display_name__ = "Matplotlib Canvas"

register_mpl_widget()



from __future__ import annotations
from pathlib import Path
from typing import Any, Literal, overload

from himena.types import ClipboardDataModel
from himena.widgets import current_instance
from contextlib import suppress


def set_status_tip(text: str, duration: float = 10.0) -> None:
    """Set a status tip to the current main window for duration (second).

    This method can be safely called from any thread.

    Parameters
    ----------
    text : str
        Text to show in the status bar.
    duration : float, default 10.0
        Duration (seconds) to show the status tip.
    """

    with suppress(Exception):
        ins = current_instance()
        ins.set_status_tip(text, duration=duration)


def get_clipboard() -> ClipboardDataModel:
    """Get the current clipboard data."""
    return current_instance().clipboard


@overload
def set_clipboard(
    *,
    text: str | None = None,
    html: str | None = None,
    image: Any | None = None,
    files: list[str | Path] | None = None,
    interanal_data: Any | None = None,
) -> None: ...


@overload
def set_clipboard(model: ClipboardDataModel, /) -> None: ...


def set_clipboard(model=None, **kwargs) -> None:
    """Set data to clipboard."""
    ins = current_instance()
    if model is not None:
        if kwargs:
            raise TypeError("Cannot specify both model and keyword arguments")
        ins.clipboard = model
    else:
        ins.set_clipboard(**kwargs)
    return None


def notify(text: str, duration: float = 5.0) -> None:
    """Show a notification popup in the bottom right corner.

    Parameters
    ----------
    text : str
        Text to show in the notification.
    duration : float, default 5.0
        Duration (seconds) to show the notification.
    """
    ins = current_instance()
    ins.show_notification(text, duration)


def show_tooltip(
    text: str,
    duration: float = 3.0,
    behavior: Literal["stay", "follow", "until_move"] = "follow",
) -> None:
    """Show a tooltip next to the cursor for a duration (sec).

    Parameters
    ----------
    text : str
        HTML text to show in the tooltip.
    duration : float, default 3.0
        Duration (seconds) to show the tooltip.
    behavior : str, default "follow"
        Behavior of the tooltip. "stay" to show at the position where it is created,
        "follow" to follow the cursor, "until_move" to show until the cursor moves.
    """
    with suppress(Exception):
        ins = current_instance()
        ins.show_tooltip(text, duration=duration, behavior=behavior)


def append_result(item: dict[str, Any], /) -> None:
    """Append a new result to the result stack."""
    ins = current_instance()
    ins._backend_main_window._append_result(item)



from qtpy.QtWidgets import QApplication
from pytestqt.qtbot import QtBot
from himena import MainWindow
from himena_builtins.qt.full_workflow._widget import QFullWorkflowView

def _mock_single_shot(ms, slot):
    """Mock single shot to avoid timing issues in tests."""
    slot()

def test_full_workflow_widget(qtbot: QtBot, himena_ui: MainWindow):
    widget = QFullWorkflowView(himena_ui)
    widget._single_shot = _mock_single_shot
    himena_ui.add_dock_widget(widget)
    himena_ui.add_object("abc,d")
    himena_ui.exec_action("builtins:text:change-separator", with_params={})
    assert len(widget.view.list_ids()) == 2
    himena_ui.tabs[0].pop(1)
    QApplication.processEvents()
    assert len(widget.view.list_ids()) == 1



from __future__ import annotations

import logging
import traceback
from importlib import import_module
from typing import TYPE_CHECKING, Callable
from pathlib import Path
from timeit import default_timer as timer
from app_model.types import KeyBindingRule, KeyBinding
from dataclasses import dataclass, field
from himena.plugins.config import config_field

if TYPE_CHECKING:
    from himena.widgets import MainWindow
    from himena.profile import AppProfile
    from himena._app_model import HimenaApplication
    from himena.plugins.widget_class import PluginConfigType

_LOGGER = logging.getLogger(__name__)


def install_plugins(
    app: HimenaApplication, plugins: list[str]
) -> list[PluginInstallResult]:
    """Install plugins to the application."""
    from himena.plugins import AppActionRegistry
    from himena.profile import load_app_profile

    reg = AppActionRegistry.instance()
    results = []
    show_import_time = app.attributes.get("print_import_time", False)
    if show_import_time:
        print("==================")
        print("Plugin import time")
        print("==================")
    for name in plugins:
        if name in reg._installed_plugins:
            continue
        if install_result := _install_one(name, show_import_time):
            results.append(install_result)
    reg.install_to(app)
    reg._installed_plugins.extend(plugins)
    prof = load_app_profile(app.name)

    for cfg_key, cfg in reg._plugin_default_configs.items():
        cfg_dict = cfg.as_dict()
        cfg_dict_old = prof.plugin_configs.get(cfg_key, {})
        # NOTE: if same config was updated during development, it may not have some keys
        for k, v in cfg_dict.items():
            if k not in cfg_dict_old:
                cfg_dict_old[k] = v
        prof.plugin_configs[cfg_key] = cfg_dict_old

    prof.save()
    return results


def _install_one(name: str, show_import_time: bool) -> PluginInstallResult | None:
    _time_0 = timer()
    _exc = None
    if isinstance(name, str):
        if name.endswith(".py"):
            if not Path(name).exists():
                _LOGGER.error(
                    f"Plugin file {name} does not exists but is listed in the "
                    "application profile."
                )
                return None
            import runpy

            mod_namespace = runpy.run_path(name)
        else:
            try:
                mod = import_module(name)
            except ModuleNotFoundError:
                _LOGGER.error(
                    f"Plugin module {name} is not installed but is listed in the "
                    "application profile."
                )
                return None
            except Exception as e:
                msg = "".join(traceback.format_exception(type(e), e, e.__traceback__))
                _LOGGER.error(
                    f"Error installing plugin {name}, traceback follows:\n{msg}"
                )
                _exc = e
            mod_namespace = vars(mod)
    else:
        raise TypeError(f"Invalid plugin type: {type(name)}")

    startup_callback = mod_namespace.get("on_himena_startup", None)

    _msec = (timer() - _time_0) * 1000
    if show_import_time and _exc is None:
        color = _color_for_time(_msec)
        print(f"{color}{name}\t{_msec:.3f} msec\033[0m")
    return PluginInstallResult(name, _msec, _exc, startup=startup_callback)


def override_keybindings(app: HimenaApplication, prof: AppProfile) -> None:
    """Override keybindings in the application."""
    for ko in prof.keybinding_overrides:
        kb_int = KeyBinding.from_str(ko.to_normed_str()).to_int()
        app.keybindings._keymap.pop(kb_int, None)
        app.keybindings.register_keybinding_rule(
            ko.command_id,
            KeyBindingRule(primary=ko.to_normed_str()),
        )


@dataclass
class GlobalConfig:
    num_recent_files_to_show: int = config_field(default=10)
    num_recent_sessions_to_show: int = config_field(default=3)
    subwindow_bar_height: int = config_field(default=18, min=10, max=45)


def install_default_configs() -> None:
    register_config("himena", "Global Settings", GlobalConfig)


def register_config(plugin_id: str, title: str, cfg: PluginConfigType) -> None:
    """Register a plugin-specific configuration.

    Registered options can be accessed by the `get_config` function. Note that a plugin
    configuration is not used as a global variable. If changing the config leads to a
    different output, it is not what we expect. What a plugin config usually does is
    "path to some-app executable", "cache directory", etc.

    Parameters
    ----------
    plugin_id : str
        The unique identifier for the plugin.
    title : str
        The title of the configuration shown in the setting dialog.
    cfg : PluginConfigType
        The configuration class (dict, dataclass or pydantic.BaseModel) that defines
        the plugin's settings.

    Examples
    --------
    ```python
    from dataclasses import dataclass
    from himena.plugins import register_config, config_field

    @dataclass
    class MyPluginConfig:
        my_option: str = config_field(default="default_value")

    register_config("my-plugin-id", "My Plugin Settings", MyPluginConfig)
    ```

    """
    from himena.plugins.actions import AppActionRegistry, PluginConfigTuple

    reg = AppActionRegistry.instance()
    tup = PluginConfigTuple(title, cfg, type(cfg))
    cfg_dict = tup.as_dict()
    for key, value_dict in cfg_dict.items():
        assert isinstance(value_dict, dict)
        if value_dict.get("value") is None and "choices" not in value_dict:
            # Unlike using magicgui in runtime, no annotations can be saved to the
            # config field.
            raise ValueError(
                f"Key {key!r} ofr config {cfg!r} must have a default value or choices. "
                f"Equivalent dict was:\n{cfg_dict!r}"
            )
    reg._plugin_default_configs[plugin_id] = PluginConfigTuple(title, cfg, type(cfg))


@dataclass
class PluginInstallResult:
    plugin: str
    time: float = field(default=0.0)
    error: Exception | None = None
    startup: Callable[[MainWindow], None] | None = None

    def call_startup(self, ui: MainWindow):
        if self.startup is not None:
            try:
                self.startup(ui)
            except Exception as e:
                _LOGGER.error(
                    f"Error calling on_himena_startup of plugin {self.plugin}: {e!r}"
                )


def _color_for_time(msec: float) -> str:
    """Return a color code for the given time in milliseconds."""
    if msec < 80:
        return "\033[92m"  # green
    elif msec < 700:
        return "\033[93m"  # yellow
    else:
        return "\033[91m"  # red



"""Data wrappers for DataFrame and Array."""

from himena.data_wrappers._array import wrap_array, ArrayWrapper
from himena.data_wrappers._dataframe import (
    wrap_dataframe,
    DataFrameWrapper,
    list_installed_dataframe_packages,
    read_csv,
)

__all__ = [
    "wrap_dataframe",
    "DataFrameWrapper",
    "list_installed_dataframe_packages",
    "read_csv",
    "wrap_array",
    "ArrayWrapper",
]



# pragma: no cover
import tempfile
import warnings
import pytest
import gc
from pathlib import Path
from qtpy.QtWidgets import QApplication
from app_model import Application
from pytestqt.qtbot import QtBot


@pytest.fixture(scope="session", autouse=True)
def patch_user_data_dir(request: pytest.FixtureRequest):
    from himena.profile import patch_user_data_dir

    with tempfile.TemporaryDirectory() as tmpdir:
        with patch_user_data_dir(tmpdir):
            yield


@pytest.fixture
def himena_ui(qtbot: QtBot, request: pytest.FixtureRequest):
    _factory = _make_himena_ui(qtbot, request)
    window = next(_factory)()
    yield window


@pytest.fixture
def make_himena_ui(qtbot: QtBot, request: pytest.FixtureRequest):
    yield from _make_himena_ui(qtbot, request)


def _make_himena_ui(qtbot: QtBot, request: pytest.FixtureRequest):
    from himena import new_window, MainWindow
    from himena._app_model._application import HimenaApplication
    from himena.widgets._initialize import _APP_INSTANCES, cleanup

    if _APP_INSTANCES:
        existing = []
        for ins in _APP_INSTANCES.values():
            for each in ins:
                pytest_name = getattr(each, "_pytest_name", "None")
                existing.append(f"{each} ({pytest_name})")
        existing_str = "    \n".join(existing)
        cleanup()
        warnings.warn(
            f"Instances not cleaned up in the previous session.\n"
            f"Existing instances:\n    {existing_str}",
            RuntimeWarning,
            stacklevel=2,
        )
    window: MainWindow | None = None

    def _factory(backend="qt"):
        nonlocal window
        window = new_window(backend=backend)
        window._instructions = window._instructions.updated(
            confirm=False,
            file_dialog_response=_raise_dialog_error,
            choose_one_dialog_response=_raise_dialog_error,
        )
        window._pytest_name = request.node.name
        if backend == "qt":
            qtbot.add_widget(window._backend_main_window)
        return window

    try:
        yield _factory
    finally:
        assert window is not None
        Application.destroy(window.model_app)
        Application.destroy(".")
        window.close()
        assert window.model_app not in Application._instances
        assert window.model_app not in HimenaApplication._instances
        assert len(_APP_INSTANCES) == 0

        QApplication.processEvents()
        QApplication.processEvents()
        QApplication.processEvents()
        gc.collect()


@pytest.fixture
def sample_dir() -> Path:
    return Path(__file__).parent.parent.parent / "tests" / "samples"


class UserResponseRequested(RuntimeError):
    """Raised when a user response is requested during testing."""


def _raise_dialog_error():  # pragma: no cover
    raise UserResponseRequested(
        "User response is requested during testing. This error is raised to prevent "
        "blocking the test execution. To fix this, you can use the testing functions "
        "`file_dialog_response` and `choose_one_dialog_response` from the "
        "`himena.testing` submodule."
    )



from contextlib import contextmanager
import json
from pathlib import Path
from typing import Any, Iterable
import warnings
from platformdirs import user_data_dir
from pydantic import BaseModel, Field, field_validator
import importlib.metadata
from packaging.version import Version
from himena.consts import ALLOWED_LETTERS


USER_DATA_DIR = Path(user_data_dir("himena"))


@contextmanager
def patch_user_data_dir(path: str | Path):
    """Change the user data directory to avoid pytest updates the local state."""
    global USER_DATA_DIR
    old = USER_DATA_DIR
    USER_DATA_DIR = Path(path)
    try:
        yield
    finally:
        USER_DATA_DIR = old


def data_dir() -> Path:
    """Get the user data directory."""
    if not USER_DATA_DIR.exists():
        USER_DATA_DIR.mkdir(parents=True)
    return USER_DATA_DIR


def profile_dir() -> Path:
    _dir = data_dir() / "profiles"
    if not _dir.exists():
        _dir.mkdir(parents=True)
    return _dir


# Plugin place and the version added.
# NOTE: The version that added this `DEFAULT_PLUGINS` variable is 0.0.8.
DEFAULT_PLUGINS: list[tuple[str, Version]] = [
    # Dock widgets
    ("himena_builtins.qt.console", Version("0.0.0")),
    ("himena_builtins.qt.explorer", Version("0.0.0")),
    ("himena_builtins.qt.favorites", Version("0.0.9")),
    ("himena_builtins.qt.history", Version("0.0.0")),
    ("himena_builtins.qt.output", Version("0.0.0")),
    ("himena_builtins.qt.full_workflow", Version("0.0.13")),
    # Matplotlib
    ("himena_builtins.qt.plot", Version("0.0.0")),
    # Widgets
    ("himena_builtins.qt.array", Version("0.0.8")),
    ("himena_builtins.qt.basic", Version("0.0.8")),
    ("himena_builtins.qt.dataframe", Version("0.0.8")),
    ("himena_builtins.qt.image", Version("0.0.8")),
    ("himena_builtins.qt.ipynb", Version("0.0.8")),
    ("himena_builtins.qt.rois", Version("0.0.8")),
    ("himena_builtins.qt.stack", Version("0.0.8")),
    ("himena_builtins.qt.table", Version("0.0.8")),
    ("himena_builtins.qt.text", Version("0.0.8")),
    # Tools
    ("himena_builtins.tools.array", Version("0.0.8")),
    ("himena_builtins.tools.conversions", Version("0.0.8")),
    ("himena_builtins.tools.dataframe", Version("0.0.8")),
    ("himena_builtins.tools.dict", Version("0.0.8")),
    ("himena_builtins.tools.image", Version("0.0.8")),
    ("himena_builtins.tools.others", Version("0.0.8")),
    ("himena_builtins.tools.plot", Version("0.0.8")),
    ("himena_builtins.tools.table", Version("0.0.8")),
    ("himena_builtins.tools.text", Version("0.0.8")),
    # Others
    ("himena_builtins.io", Version("0.0.0")),
    ("himena_builtins.new", Version("0.0.0")),
    ("himena_builtins.user_modifications", Version("0.0.8")),
]


def _default_plugins() -> list[str]:
    """Factory function for the default plugin list."""
    return [place for place, _ in DEFAULT_PLUGINS]


def _current_version() -> str:
    """Get the current version of himena."""
    try:
        return importlib.metadata.version("himena")
    except Exception:
        return "0.0.1"


class KeyBindingOverride(BaseModel):
    key: str
    command_id: str

    def to_normed_str(self) -> str:
        return self.key.replace(", ", " ")


class AppProfile(BaseModel):
    """Model of a profile."""

    name: str = Field(
        default="default",
        description="Name of the profile.",
        frozen=True,
    )
    version: str = Field(
        default_factory=_current_version,
        description="Version of this profile created.",
        frozen=True,
    )
    plugins: list[str] = Field(
        default_factory=_default_plugins, description="List of plugins to load."
    )
    theme: str = Field(default="light-green", description="Theme to use.")
    startup_commands: list[tuple[str, dict[str, Any] | None]] = Field(
        default_factory=list,
        description="Startup commands that will be executed when the app starts.",
    )
    keybinding_overrides: list[KeyBindingOverride] = Field(default_factory=list)
    plugin_configs: dict[str, dict[str, Any]] = Field(default_factory=dict)

    @classmethod
    def from_json(cls, path) -> "AppProfile":
        """Construct an AppProfile from a json file."""
        with open(path) as f:
            data = json.load(f)
        version_saved = Version(data.get("version", "0.0.1"))
        self = cls(**data)
        if version_saved < Version(_current_version()):
            for place, version_added in DEFAULT_PLUGINS:
                if version_added > version_saved and place not in self.plugins:
                    # Add the default plugin that is implemented after the profile was
                    # saved.
                    self.plugins.append(place)
        return self

    @classmethod
    def default(cls, save: bool = False) -> "AppProfile":
        """Return the default profile."""
        prof = AppProfile()
        if save and not (profile_dir() / f"{prof.name}.json").exists():
            prof.save()
        return prof

    def save(self, path: str | Path | None = None) -> None:
        """Save profile as a json file."""
        if path is None:
            path = self.profile_path()
        json_string = json.dumps(self.model_dump(), indent=4)
        with open(path, "w") as f:
            f.write(json_string)
        return None

    def profile_path(self) -> Path:
        """Path to this profile."""
        return profile_dir() / f"{self.name}.json"

    def with_name(self, name: str) -> "AppProfile":
        """Return a new profile with a new name."""
        return self.model_copy(update={"name": name})

    def with_plugins(self, plugins: list[str]) -> "AppProfile":
        """Return a new profile with new plugins."""
        return self.model_copy(update={"plugins": plugins})

    def with_plugin_configs(self, configs: dict[str, dict[str, Any]]) -> "AppProfile":
        """Return a new profile with new plugin configs."""
        return self.model_copy(update={"plugin_configs": configs})

    def with_keybinding_override(self, key: str, command_id: str) -> "AppProfile":
        """Return a new profile with new keybind overrides."""
        _overrides = self.keybinding_overrides.copy()
        for entry in _overrides:
            if entry.command_id == command_id:
                if key:
                    entry.key = key
                else:
                    _overrides.remove(entry)
                break
        else:
            if key:
                _overrides.append(KeyBindingOverride(key=key, command_id=command_id))
        return self.model_copy(update={"keybinding_overrides": _overrides})

    def update_plugin_config(self, plugin_id: str, **kwargs) -> None:
        """Update the config of the plugin specified by `plugin_id`"""
        from himena.plugins import AppActionRegistry
        from himena.plugins.widget_plugins import WidgetCallbackBase

        reg = AppActionRegistry.instance()
        configs = self.plugin_configs.copy()
        # NOTE: during development, keys of cur_config and configs[plugin_id] may
        # differ. `cur_config` has all the keys that should exist in the current
        # implementation.
        cur_config = reg._plugin_default_configs[plugin_id].as_dict()
        if plugin_id in configs:
            # Profile already has the plugin config
            for ckey, cval in configs[plugin_id].items():
                if ckey in cur_config:
                    cur_config[ckey] = cval
        for k, v in kwargs.items():
            if k in cur_config:
                cur_config[k]["value"] = v
        configs[plugin_id] = cur_config
        self.with_plugin_configs(configs).save()

        # update existing dock widgets with the new config
        params = {}
        for key, opt in cur_config.items():
            params[key] = opt["value"]
        if cb := WidgetCallbackBase.instance_for_command_id(plugin_id):
            for dock in cb._all_widgets:
                # the internal widget should always has the method "update_configs"
                dock.update_configs(params)

    @field_validator("name")
    def _validate_name(cls, value):
        # check if value is a valid file name
        if not all(c in ALLOWED_LETTERS for c in value):
            raise ValueError(f"Invalid profile name: {value}")
        return value


def load_app_profile(name: str, *, create_default: bool = False) -> AppProfile:
    path = profile_dir() / f"{name}.json"
    if path.exists():
        try:
            prof = AppProfile.from_json(path)
        except json.JSONDecodeError as e:
            raise ValueError(
                f"Could not load profile {name!r}. Probably the file is corrupted. "
                f"Please manually fix {path}.\nOriginal error ... \n\n"
                f"{type(e).__name__}: {e}",
            ) from None
        return prof
    if create_default:
        AppProfile.default().with_name(name).save(path)
        return AppProfile.from_json(path)
    raise ValueError(
        f"Profile {name!r} does not exist. Please create a new profile with:\n"
        f"$ himena --new {name}"
    )


def iter_app_profiles() -> Iterable[AppProfile]:
    for path in profile_dir().glob("*.json"):
        try:
            yield AppProfile.from_json(path)
        except Exception:
            warnings.warn(f"Could not load profile {path}.")


def new_app_profile(name: str) -> AppProfile:
    """Create a new profile."""
    path = profile_dir() / f"{name}.json"
    if path.exists():
        raise ValueError(f"Profile {name!r} already exists.")
    profile = AppProfile.default().with_name(name)
    profile.save(path)
    return profile


def remove_app_profile(name: str) -> None:
    """Remove an existing profile."""
    path = profile_dir() / f"{name}.json"
    return path.unlink()



from typing import TYPE_CHECKING
from app_model.expressions import ContextKey, ContextNamespace
from himena.types import WindowState
from himena._utils import get_widget_class_id, get_user_context

if TYPE_CHECKING:
    from himena.widgets import MainWindow


def _is_active_window_savable(ui: "MainWindow") -> bool:
    if win := ui.current_window:
        return win.supports_to_model
    return False


def _is_active_window_editable(ui: "MainWindow") -> bool:
    if win := ui.current_window:
        return win.is_editable
    return False


def _is_active_window_track_modification(ui: "MainWindow") -> bool:
    if win := ui.current_window:
        return win._data_modifications.track_enabled
    return False


def _active_window_state(ui: "MainWindow"):
    if win := ui.current_window:
        return win.state
    return WindowState.NORMAL


def _is_active_window_focused(ui: "MainWindow") -> bool:
    return ui.current_window is not None


def _num_sub_windows(ui: "MainWindow") -> int:
    if area := ui.tabs.current():
        return area.len()
    return 0


def _num_tabs(ui: "MainWindow") -> int:
    return ui.tabs.len()


def _get_model_types(ui: "MainWindow") -> str | None:
    try:
        if (win := ui.current_window) and win.supports_to_model:
            out = win.model_type()
            if out is None:
                out = win.to_model().type
                if out is None:
                    return None
            return out.split(".")
    except Exception:
        return None
    return None


def _active_window_model_type(ui: "MainWindow") -> str | None:
    if out := _get_model_types(ui):
        return out[0]
    return None


def _active_window_model_subtype_1(ui: "MainWindow") -> str | None:
    if out := _get_model_types(ui):
        if len(out) < 2:
            return None
        return out[1]
    return None


def _active_window_model_subtype_2(ui: "MainWindow") -> str | None:
    if out := _get_model_types(ui):
        if len(out) < 3:
            return None
        return out[2]
    return None


def _active_window_model_subtype_3(ui: "MainWindow") -> str | None:
    if out := _get_model_types(ui):
        if len(out) < 4:
            return None
        return out[3]
    return None


def _active_window_widget_id(ui: "MainWindow") -> str | None:
    if win := ui.current_window:
        return get_widget_class_id(type(win.widget))
    return None


def _get_user_context(ui: "MainWindow") -> dict | None:
    if area := ui.tabs.current():
        if win := area.current():
            return get_user_context(win.widget)
    return None


class AppContext(ContextNamespace["MainWindow"]):
    """Context namespace for the himena main window."""

    is_active_window_supports_to_model = ContextKey(
        False,
        "if the current window supports to_model method",
        _is_active_window_savable,
    )
    is_active_window_editable = ContextKey(
        False,
        "if the current window is editable",
        _is_active_window_editable,
    )
    is_active_window_track_modification = ContextKey(
        False,
        "if the current window is tracking user modifications",
        _is_active_window_track_modification,
    )
    active_window_state = ContextKey(
        WindowState.NORMAL,
        "state of the sub-window",
        _active_window_state,
    )
    is_subwindow_focused = ContextKey(
        False,
        "if a sub-window is focused",
        _is_active_window_focused,
    )
    num_tabs = ContextKey(
        0,
        "number of tabs",
        _num_tabs,
    )
    num_sub_windows = ContextKey(
        0,
        "number of sub-windows in the current tab",
        _num_sub_windows,
    )
    active_window_model_type = ContextKey(
        None,
        "type of the model of the active window",
        _active_window_model_type,
    )
    active_window_model_subtype_1 = ContextKey(
        None,
        "subtype of the model of the active window",
        _active_window_model_subtype_1,
    )
    active_window_model_subtype_2 = ContextKey(
        None,
        "subtype of the model of the active window",
        _active_window_model_subtype_2,
    )
    active_window_model_subtype_3 = ContextKey(
        None,
        "subtype of the model of the active window",
        _active_window_model_subtype_3,
    )
    active_window_widget_id = ContextKey(
        None,
        "widget class id of the active window",
        _active_window_widget_id,
    )
    user_context = ContextKey(
        None,
        "User-defined context",
        _get_user_context,
    )

    def _update(self, ui):
        for k, v in self._getters.items():
            setattr(self, k, v(ui))



from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING, Iterator, Literal
from qtpy import QtWidgets as QtW, QtCore, QtGui
from superqt import QToggleSwitch

from himena.qt._qsvg import QColoredSVGIcon
from himena.workflow import RemoteReaderMethod
from himena.consts import MonospaceFontFamily, IS_WINDOWS
from himena.utils.cli import local_to_remote
from himena_builtins._consts import ICON_PATH
from himena_builtins.qt.explorer._base import (
    QBaseRemoteExplorerWidget,
    ls_args_to_items,
    exec_command,
    stat_args_to_type,
)
from himena_builtins.qt.widgets._shared import labeled

if TYPE_CHECKING:
    from himena_builtins.qt.explorer import FileExplorerSSHConfig
    from himena.qt import MainWindowQt


class QSSHRemoteExplorerWidget(QBaseRemoteExplorerWidget):
    """A widget for exploring remote files via SSH.

    This widget will execute `ls`, `ssh` and `scp` commands to list, read and send
    files when needed. This widget accepts copy-and-paste drag-and-drop from the local
    file system, including the normal explorer dock widget and the OS file explorer.

    If you are using Windows, checking the "Use WSL" switch will forward all the
    subprocess commands to WSL.
    """

    def __init__(self, ui: MainWindowQt) -> None:
        super().__init__(ui)
        font = QtGui.QFont(MonospaceFontFamily)
        self._host_edit = QtW.QLineEdit()
        self._host_edit.setFont(font)
        self._host_edit.setMaximumWidth(100)
        self._user_name_edit = QtW.QLineEdit()
        self._user_name_edit.setFont(font)
        self._user_name_edit.setMaximumWidth(80)
        self._port_edit = QtW.QLineEdit("22")
        self._port_edit.setFont(font)
        self._port_edit.setValidator(QtGui.QIntValidator(0, 65535))
        self._port_edit.setMaximumWidth(40)
        self._is_wsl_switch = QToggleSwitch()
        self._is_wsl_switch.setText("Use WSL")
        self._is_wsl_switch.setFixedHeight(24)
        self._is_wsl_switch.setChecked(False)
        self._is_wsl_switch.setVisible(IS_WINDOWS)
        self._is_wsl_switch.setToolTip(
            "Use WSL (Windows Subsystem for Linux) to access remote files. If \n "
            "checked, all the subprocess commands such as `ls` will be prefixed \n"
            "with `wsl -e`."
        )
        self._protocol_choice = QtW.QComboBox()
        self._protocol_choice.addItems(["rsync", "scp"])
        self._protocol_choice.setCurrentIndex(0)
        self._protocol_choice.setToolTip("Choose the protocol to send files.")

        self._show_hidden_files_switch = QToggleSwitch()
        self._show_hidden_files_switch.setText("Hidden Files")
        self._show_hidden_files_switch.setToolTip("Also show hidden files")
        self._show_hidden_files_switch.setFixedHeight(24)
        self._show_hidden_files_switch.setChecked(False)

        self._last_dir_btn = QtW.QPushButton("←")
        self._last_dir_btn.setFixedWidth(20)
        self._last_dir_btn.setToolTip("Back to last directory")

        self._up_one_btn = QtW.QPushButton("↑")
        self._up_one_btn.setFixedWidth(20)
        self._up_one_btn.setToolTip("Up one directory")
        self._refresh_btn = QtW.QToolButton()
        self._refresh_btn.setToolTip("Refresh current directory (F5)")

        self._conn_btn = QtW.QPushButton("Connect")
        self._conn_btn.setFixedWidth(60)
        self._conn_btn.setToolTip("Connect to the remote host with the given user name")

        layout = QtW.QVBoxLayout(self)

        hlayout0 = QtW.QHBoxLayout()
        hlayout0.setContentsMargins(0, 0, 0, 0)
        layout.addLayout(hlayout0)
        hlayout0.addWidget(labeled("Host:", self._host_edit, label_width=30), 3)
        hlayout0.addWidget(labeled("User:", self._user_name_edit, label_width=30), 2)
        hlayout0.addWidget(labeled("Port:", self._port_edit, label_width=30), 2)

        hlayout1 = QtW.QHBoxLayout()
        hlayout1.setContentsMargins(0, 0, 0, 0)
        layout.addLayout(hlayout1)
        hlayout1.addWidget(self._is_wsl_switch)
        hlayout1.addWidget(self._protocol_choice)
        hlayout1.addWidget(self._conn_btn)

        layout.addWidget(QSeparator())
        layout.addWidget(labeled("Path:", self._pwd_widget))

        hlayout2 = QtW.QHBoxLayout()
        hlayout2.setContentsMargins(0, 0, 0, 0)
        hlayout2.addWidget(self._last_dir_btn, 0, QtCore.Qt.AlignmentFlag.AlignLeft)
        hlayout2.addWidget(self._up_one_btn, 0, QtCore.Qt.AlignmentFlag.AlignLeft)
        hlayout2.addWidget(QtW.QWidget(), 100)  # spacer
        hlayout2.addWidget(self._show_hidden_files_switch)
        hlayout2.addWidget(self._refresh_btn, 0, QtCore.Qt.AlignmentFlag.AlignRight)
        layout.addLayout(hlayout2)
        layout.addWidget(self._filter_widget)
        layout.addWidget(self._file_list_widget)

        self._conn_btn.clicked.connect(lambda: self._set_current_path(Path("~")))
        self._refresh_btn.clicked.connect(self._refresh_pwd)
        self._last_dir_btn.clicked.connect(
            lambda: self._set_current_path(self._last_dir)
        )
        self._up_one_btn.clicked.connect(
            lambda: self._set_current_path(self._pwd.parent)
        )
        self._show_hidden_files_switch.toggled.connect(self._refresh_pwd)
        self._light_background = True
        self.themeChanged.connect(self._on_theme_changed)

    def _on_theme_changed(self, theme) -> None:
        color = "#222222" if self._light_background else "#eeeeee"
        self._refresh_btn.setIcon(
            QColoredSVGIcon.fromfile(ICON_PATH / "refresh.svg", color)
        )

    def _set_busy(self, busy: bool):
        self._conn_btn.setEnabled(not busy)
        self._refresh_btn.setEnabled(not busy)
        self._last_dir_btn.setEnabled(not busy)
        self._up_one_btn.setEnabled(not busy)
        self._show_hidden_files_switch.setEnabled(not busy)
        self._file_list_widget.setEnabled(not busy)
        self._pwd_widget.setEnabled(not busy)

    def _host_name(self) -> str:
        username = self._user_name_edit.text()
        host = self._host_edit.text()
        return f"{username}@{host}"

    def _make_reader_method(self, path: Path, is_dir: bool) -> RemoteReaderMethod:
        return RemoteReaderMethod(
            host=self._host_edit.text(),
            username=self._user_name_edit.text(),
            path=path,
            port=int(self._port_edit.text()),
            wsl=self._is_wsl_switch.isChecked(),
            protocol=self._protocol_choice.currentText(),
            force_directory=is_dir,
        )

    def _iter_file_items(self, path) -> Iterator[QtW.QTreeWidgetItem]:
        opt = "-lhAF" if self._show_hidden_files_switch.isChecked() else "-lhF"
        host, port = self._host_and_port()
        args = ["ssh", "-p", port, host, "ls", path + "/", opt]
        yield from ls_args_to_items(self._with_wsl_prefix(args))

    def _get_file_type(self, path: str) -> Literal["d", "f"]:
        host, port = self._host_and_port()
        args = ["ssh", "-p", port, host, "stat", path, "--format='%F'"]
        return stat_args_to_type(self._with_wsl_prefix(args))

    def _move_files(self, src: str, dst: str) -> None:
        host, port = self._host_and_port()
        args = ["ssh", "-p", port, host, "mv", src, dst]
        exec_command(self._with_wsl_prefix(args))

    def _trash_files(self, paths: list[str]) -> None:
        host, port = self._host_and_port()
        args = ["ssh", "-p", port, host, "trash", *paths]
        exec_command(self._with_wsl_prefix(args))

    def _host_and_port(self) -> tuple[str, str]:
        """Return the host and port as a tuple."""
        host = self._host_edit.text()
        port = self._port_edit.text()
        return host, str(int(port))

    def _with_wsl_prefix(self, args: list[str]) -> list[str]:
        """Prefix the command with 'wsl -e' if WSL is enabled."""
        if self._is_wsl_switch.isChecked() and IS_WINDOWS:
            return ["wsl", "-e"] + args
        return args

    def _make_reader_method_from_str(
        self, line: str, is_dir: bool
    ) -> RemoteReaderMethod:
        """Create a RemoteReaderMethod from a string path."""
        return RemoteReaderMethod.from_str(
            line,
            wsl=self._is_wsl_switch.isChecked(),
            protocol=self._protocol_choice.currentText(),
            force_directory=is_dir,
        )

    def update_configs(
        self,
        cfg: FileExplorerSSHConfig,
    ) -> None:
        self._host_edit.setText(cfg.default_host)
        self._user_name_edit.setText(cfg.default_user)
        self._port_edit.setText(str(cfg.default_port))
        self._is_wsl_switch.setChecked(cfg.default_use_wsl)
        self._protocol_choice.setCurrentText(cfg.default_protocol)
        if cfg.default_host and cfg.default_user and self._pwd == Path("~"):
            self._set_current_path(Path("~"))

    def _send_file_args(self, src: Path, dst_remote: str, is_dir: bool = False):
        return local_to_remote(
            self._protocol_choice.currentText(),
            src,
            f"{self._host_name()}:{dst_remote}",
            is_wsl=self._is_wsl_switch.isChecked(),
            is_dir=is_dir,
            port=int(self._port_edit.text()),
        )

    def _send_file(self, src: Path, dst_remote: str, is_dir: bool = False):
        exec_command(self._send_file_args(src, dst_remote, is_dir=is_dir))


class QSeparator(QtW.QFrame):
    def __init__(self):
        super().__init__()
        self.setFrameShape(QtW.QFrame.Shape.HLine)
        self.setFrameShadow(QtW.QFrame.Shadow.Sunken)
        self.setFixedHeight(2)
        self.setSizePolicy(
            QtW.QSizePolicy.Policy.Expanding, QtW.QSizePolicy.Policy.Fixed
        )



from __future__ import annotations
import argparse
from pathlib import Path
from himena.consts import ALLOWED_LETTERS


class HimenaCliNamespace(argparse.Namespace):
    """The namespace returned by the CLI parser."""

    profile: str | None
    path: str | None
    log_level: str
    new: str | None
    remove: str | None
    install: list[str]
    uninstall: list[str]
    uninstall_outdated: bool
    get: list[str]
    list_plugins: bool
    list_profiles: bool
    clear_plugin_configs: bool
    import_time: bool
    host: str = "localhost"
    port: int = 49200

    def assert_args_not_given(self) -> None:
        if self.profile is not None or self.path is not None:
            raise ValueError(
                "Profile name and file path cannot be given with this option."
            )

    def norm_profile_and_path(self) -> HimenaCliNamespace:
        if self.profile is not None and not _is_profile_name(self.profile):
            self.path = self.profile
            self.profile = None

        return self

    def abs_path(self) -> str | None:
        """Return the absolute path of the file to open."""
        if self.path is None:
            return None
        fp = Path(self.path).expanduser()
        if fp.exists():
            fp = fp.resolve()
        return str(fp)


def _is_profile_name(arg: str) -> bool:
    return all(c in ALLOWED_LETTERS for c in arg)


class HimenaArgumentParser(argparse.ArgumentParser):
    def __init__(self):
        super().__init__(prog="himena", description="Start the himena GUI application.")
        ### Configure the parser ###
        # fmt: off
        self.add_argument(
            "profile", nargs="?", default=None,
            help=(
                "Profile name. If not given, the default profile is used. If a file path "
                "is given, it will be interpreted as the next 'path' argument and this "
                "argument will be set to None."
            )
        )
        self.add_argument(
            "path", nargs="?", default=None,
            help="File path to open with the GUI."
        )
        self.add_argument(
            "--log-level", nargs="?", default="WARNING",
            choices=["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"],
            help="Set the default log level.",
        )
        self.add_argument(
            "--new", default=None,
            help="Create a new profile with the given name."
        )
        self.add_argument(
            "--remove", default=None,
            help="Remove the profile of the given name."
        )
        self.add_argument(
            "--install", nargs="+", default=[], help="Install the given plugins."
        )
        self.add_argument(
            "--uninstall", nargs="+", default=[], help="Uninstall the given plugins."
        )
        self.add_argument(
            "--uninstall-outdated", action="store_true",
            help=(
                "Uninstall all the outdated plugins (plugins that are listed in the) "
                "profile but cannot be found in the Python environment."
            )
        )
        self.add_argument(
            "--get", "-g", nargs="+", default=[],
            help="Install packages from PyPI and add them as plugins."
        )
        self.add_argument(
            "--list-plugins", action="store_true", help="List all the available plugins."
        )
        self.add_argument(
            "--list-profiles", action="store_true",
            help="List all the available profiles."
        )
        self.add_argument(
            "--clear-plugin-configs", action="store_true",
            help="Clear all the plugin configurations in the given profile."
        )
        self.add_argument(
            "--import-time", action="store_true",
            help="Print the import time of the plugins."
        )
        self.add_argument(
            "--host", type=str, default="localhost",
            help="Socket host name to use for the GUI.",
        )
        self.add_argument(
            "--port", type=int, default=49200,
            help="Socket port number to use for the GUI.",
        )
        # fmt: on

    def parse_args(self, args=None, namespace=None) -> HimenaCliNamespace:
        args = super().parse_args(args, namespace)
        return HimenaCliNamespace(**vars(args))
